# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .. import errors
from .. import graph as _mod_graph
from .. import tests
from ..revision import NULL_REVISION
from . import TestCaseWithMemoryTransport

# Ancestry 1:
#
#  NULL_REVISION
#       |
#     rev1
#      /\
#  rev2a rev2b
#     |    |
#   rev3  /
#     |  /
#   rev4
ancestry_1 = {b'rev1': [NULL_REVISION], b'rev2a': [b'rev1'], b'rev2b': [b'rev1'],
              b'rev3': [b'rev2a'], b'rev4': [b'rev3', b'rev2b']}


# Ancestry 2:
#
#  NULL_REVISION
#    /    \
# rev1a  rev1b
#   |
# rev2a
#   |
# rev3a
#   |
# rev4a
ancestry_2 = {b'rev1a': [NULL_REVISION], b'rev2a': [b'rev1a'],
              b'rev1b': [NULL_REVISION], b'rev3a': [b'rev2a'], b'rev4a': [b'rev3a']}


# Criss cross ancestry
#
#     NULL_REVISION
#         |
#        rev1
#        /  \
#    rev2a  rev2b
#       |\  /|
#       |  X |
#       |/  \|
#    rev3a  rev3b
criss_cross = {b'rev1': [NULL_REVISION], b'rev2a': [b'rev1'], b'rev2b': [b'rev1'],
               b'rev3a': [b'rev2a', b'rev2b'], b'rev3b': [b'rev2b', b'rev2a']}


# Criss-cross 2
#
#  NULL_REVISION
#    /   \
# rev1a  rev1b
#   |\   /|
#   | \ / |
#   |  X  |
#   | / \ |
#   |/   \|
# rev2a  rev2b
criss_cross2 = {b'rev1a': [NULL_REVISION], b'rev1b': [NULL_REVISION],
                b'rev2a': [b'rev1a', b'rev1b'], b'rev2b': [b'rev1b', b'rev1a']}


# Mainline:
#
#  NULL_REVISION
#       |
#      rev1
#      /  \
#      | rev2b
#      |  /
#     rev2a
mainline = {b'rev1': [NULL_REVISION], b'rev2a': [b'rev1', b'rev2b'],
            b'rev2b': [b'rev1']}


# feature branch:
#
#  NULL_REVISION
#       |
#      rev1
#       |
#     rev2b
#       |
#     rev3b
feature_branch = {b'rev1': [NULL_REVISION],
                  b'rev2b': [b'rev1'], b'rev3b': [b'rev2b']}


# History shortcut
#  NULL_REVISION
#       |
#     rev1------
#     /  \      \
#  rev2a rev2b rev2c
#    |  /   \   /
#  rev3a    rev3b
history_shortcut = {b'rev1': [NULL_REVISION], b'rev2a': [b'rev1'],
                    b'rev2b': [b'rev1'], b'rev2c': [b'rev1'],
                    b'rev3a': [b'rev2a', b'rev2b'], b'rev3b': [b'rev2b', b'rev2c']}

# Extended history shortcut
#  NULL_REVISION
#       |
#       a
#       |\
#       b |
#       | |
#       c |
#       | |
#       d |
#       |\|
#       e f
extended_history_shortcut = {b'a': [NULL_REVISION],
                             b'b': [b'a'],
                             b'c': [b'b'],
                             b'd': [b'c'],
                             b'e': [b'd'],
                             b'f': [b'a', b'd'],
                             }

# Double shortcut
# Both sides will see b'A' first, even though it is actually a decendent of a
# different common revision.
#
#  NULL_REVISION
#       |
#       a
#      /|\
#     / b \
#    /  |  \
#   |   c   |
#   |  / \  |
#   | d   e |
#   |/     \|
#   f       g

double_shortcut = {b'a': [NULL_REVISION], b'b': [b'a'], b'c': [b'b'],
                   b'd': [b'c'], b'e': [b'c'], b'f': [b'a', b'd'],
                   b'g': [b'a', b'e']}

# Complex shortcut
# This has a failure mode in that a shortcut will find some nodes in common,
# but the common searcher won't have time to find that one branch is actually
# in common. The extra nodes at the beginning are because we want to avoid
# walking off the graph. Specifically, node G should be considered common, but
# is likely to be seen by M long before the common searcher finds it.
#
# NULL_REVISION
#     |
#     a
#     |
#     b
#     |
#     c
#     |
#     d
#     |\
#     e f
#     | |\
#     | g h
#     |/| |
#     i j |
#     | | |
#     | k |
#     | | |
#     | l |
#     |/|/
#     m n
complex_shortcut = {b'a': [NULL_REVISION], b'b': [b'a'], b'c': [b'b'], b'd': [b'c'],
                    b'e': [b'd'], b'f': [b'd'], b'g': [b'f'], b'h': [b'f'],
                    b'i': [b'e', b'g'], b'j': [b'g'], b'k': [b'j'],
                    b'l': [b'k'], b'm': [b'i', b'l'], b'n': [b'l', b'h']}

# NULL_REVISION
#     |
#     a
#     |
#     b
#     |
#     c
#     |
#     d
#     |\
#     e |
#     | |
#     f |
#     | |
#     g h
#     | |\
#     i | j
#     |\| |
#     | k |
#     | | |
#     | l |
#     | | |
#     | m |
#     | | |
#     | n |
#     | | |
#     | o |
#     | | |
#     | p |
#     | | |
#     | q |
#     | | |
#     | r |
#     | | |
#     | s |
#     | | |
#     |/|/
#     t u
complex_shortcut2 = {b'a': [NULL_REVISION], b'b': [b'a'], b'c': [b'b'], b'd': [b'c'],
                     b'e': [b'd'], b'f': [b'e'], b'g': [b'f'], b'h': [b'd'], b'i': [b'g'],
                     b'j': [b'h'], b'k': [b'h', b'i'], b'l': [b'k'], b'm': [b'l'], b'n': [b'm'],
                     b'o': [b'n'], b'p': [b'o'], b'q': [b'p'], b'r': [b'q'], b's': [b'r'],
                     b't': [b'i', b's'], b'u': [b's', b'j'],
                     }

# Graph where different walkers will race to find the common and uncommon
# nodes.
#
# NULL_REVISION
#     |
#     a
#     |
#     b
#     |
#     c
#     |
#     d
#     |\
#     e k
#     | |
#     f-+-p
#     | | |
#     | l |
#     | | |
#     | m |
#     | |\|
#     g n q
#     |\| |
#     h o |
#     |/| |
#     i r |
#     | | |
#     | s |
#     | | |
#     | t |
#     | | |
#     | u |
#     | | |
#     | v |
#     | | |
#     | w |
#     | | |
#     | x |
#     | |\|
#     | y z
#     |/
#     j
#
# x is found to be common right away, but is the start of a long series of
# common commits.
# o is actually common, but the i-j shortcut makes it look like it is actually
# unique to j at first, you have to traverse all of x->o to find it.
# q,m gives the walker from j a common point to stop searching, as does p,f.
# k-n exists so that the second pass still has nodes that are worth searching,
# rather than instantly cancelling the extra walker.

racing_shortcuts = {b'a': [NULL_REVISION], b'b': [b'a'], b'c': [b'b'], b'd': [b'c'],
                    b'e': [b'd'], b'f': [b'e'], b'g': [b'f'], b'h': [b'g'], b'i': [b'h', b'o'], b'j': [b'i', b'y'],
                    b'k': [b'd'], b'l': [b'k'], b'm': [b'l'], b'n': [b'm'], b'o': [b'n', b'g'], b'p': [b'f'],
                    b'q': [b'p', b'm'], b'r': [b'o'], b's': [b'r'], b't': [b's'], b'u': [b't'], b'v': [b'u'],
                    b'w': [b'v'], b'x': [b'w'], b'y': [b'x'], b'z': [b'x', b'q']}


# A graph with multiple nodes unique to one side.
#
# NULL_REVISION
#     |
#     a
#     |
#     b
#     |
#     c
#     |
#     d
#     |\
#     e f
#     |\ \
#     g h i
#     |\ \ \
#     j k l m
#     | |/ x|
#     | n o p
#     | |/  |
#     | q   |
#     | |   |
#     | r   |
#     | |   |
#     | s   |
#     | |   |
#     | t   |
#     | |   |
#     | u   |
#     | |   |
#     | v   |
#     | |   |
#     | w   |
#     | |   |
#     | x   |
#     |/ \ /
#     y   z
#

multiple_interesting_unique = {b'a': [NULL_REVISION], b'b': [b'a'], b'c': [b'b'],
                               b'd': [b'c'], b'e': [b'd'], b'f': [b'd'], b'g': [b'e'], b'h': [b'e'], b'i': [b'f'],
                               b'j': [b'g'], b'k': [b'g'], b'l': [b'h'], b'm': [b'i'], b'n': [b'k', b'l'],
                               b'o': [b'm'], b'p': [b'm', b'l'], b'q': [b'n', b'o'], b'r': [b'q'], b's': [b'r'],
                               b't': [b's'], b'u': [b't'], b'v': [b'u'], b'w': [b'v'], b'x': [b'w'],
                               b'y': [b'j', b'x'], b'z': [b'x', b'p']}


# Shortcut with extra root
# We have a long history shortcut, and an extra root, which is why we can't
# stop searchers based on seeing NULL_REVISION
#  NULL_REVISION
#       |   |
#       a   |
#       |\  |
#       b | |
#       | | |
#       c | |
#       | | |
#       d | g
#       |\|/
#       e f
shortcut_extra_root = {b'a': [NULL_REVISION],
                       b'b': [b'a'],
                       b'c': [b'b'],
                       b'd': [b'c'],
                       b'e': [b'd'],
                       b'f': [b'a', b'd', b'g'],
                       b'g': [NULL_REVISION],
                       }

#  NULL_REVISION
#       |
#       f
#       |
#       e
#      / \
#     b   d
#     | \ |
#     a   c

boundary = {b'a': [b'b'], b'c': [b'b', b'd'], b'b': [b'e'], b'd': [b'e'], b'e': [b'f'],
            b'f': [NULL_REVISION]}


# A graph that contains a ghost
#  NULL_REVISION
#       |
#       f
#       |
#       e   g
#      / \ /
#     b   d
#     | \ |
#     a   c

with_ghost = {b'a': [b'b'], b'c': [b'b', b'd'], b'b': [b'e'], b'd': [b'e', b'g'],
              b'e': [b'f'], b'f': [NULL_REVISION], NULL_REVISION: ()}

# A graph that shows we can shortcut finding revnos when reaching them from the
# side.
#  NULL_REVISION
#       |
#       a
#       |
#       b
#       |
#       c
#       |
#       d
#       |
#       e
#      / \
#     f   g
#     |
#     h
#     |
#     i

with_tail = {b'a': [NULL_REVISION], b'b': [b'a'], b'c': [b'b'], b'd': [b'c'], b'e': [b'd'],
             b'f': [b'e'], b'g': [b'e'], b'h': [b'f'], b'i': [b'h']}


class InstrumentedParentsProvider:

    def __init__(self, parents_provider):
        self.calls = []
        self._real_parents_provider = parents_provider
        get_cached = getattr(parents_provider, 'get_cached_parent_map', None)
        if get_cached is not None:
            # Only expose the underlying 'get_cached_parent_map' function if
            # the wrapped provider has it.
            self.get_cached_parent_map = self._get_cached_parent_map

    def get_parent_map(self, nodes):
        self.calls.extend(nodes)
        return self._real_parents_provider.get_parent_map(nodes)

    def _get_cached_parent_map(self, nodes):
        self.calls.append(('cached', sorted(nodes)))
        return self._real_parents_provider.get_cached_parent_map(nodes)


class SharedInstrumentedParentsProvider:

    def __init__(self, parents_provider, calls, info):
        self.calls = calls
        self.info = info
        self._real_parents_provider = parents_provider
        get_cached = getattr(parents_provider, 'get_cached_parent_map', None)
        if get_cached is not None:
            # Only expose the underlying 'get_cached_parent_map' function if
            # the wrapped provider has it.
            self.get_cached_parent_map = self._get_cached_parent_map

    def get_parent_map(self, nodes):
        self.calls.append((self.info, sorted(nodes)))
        return self._real_parents_provider.get_parent_map(nodes)

    def _get_cached_parent_map(self, nodes):
        self.calls.append((self.info, 'cached', sorted(nodes)))
        return self._real_parents_provider.get_cached_parent_map(nodes)


class TestGraphBase(tests.TestCase):

    def make_graph(self, ancestors):
        return _mod_graph.Graph(_mod_graph.DictParentsProvider(ancestors))

    def make_breaking_graph(self, ancestors, break_on):
        """Make a Graph that raises an exception if we hit a node."""
        g = self.make_graph(ancestors)
        orig_parent_map = g.get_parent_map

        def get_parent_map(keys):
            bad_keys = set(keys).intersection(break_on)
            if bad_keys:
                self.fail('key(s) {} was accessed'.format(sorted(bad_keys)))
            return orig_parent_map(keys)
        g.get_parent_map = get_parent_map
        return g


class TestGraph(TestCaseWithMemoryTransport):

    def make_graph(self, ancestors):
        return _mod_graph.Graph(_mod_graph.DictParentsProvider(ancestors))

    def prepare_memory_tree(self, location):
        tree = self.make_branch_and_memory_tree(location)
        tree.lock_write()
        tree.add('.')
        return tree

    def build_ancestry(self, tree, ancestors):
        """Create an ancestry as specified by a graph dict

        :param tree: A tree to use
        :param ancestors: a dict of {node: [node_parent, ...]}
        """
        pending = [NULL_REVISION]
        descendants = {}
        for descendant, parents in ancestors.items():
            for parent in parents:
                descendants.setdefault(parent, []).append(descendant)
        while len(pending) > 0:
            cur_node = pending.pop()
            for descendant in descendants.get(cur_node, []):
                if tree.branch.repository.has_revision(descendant):
                    continue
                parents = [p for p in ancestors[descendant] if p is not
                           NULL_REVISION]
                if len([p for p in parents if not
                        tree.branch.repository.has_revision(p)]) > 0:
                    continue
                tree.set_parent_ids(parents)
                if len(parents) > 0:
                    left_parent = parents[0]
                else:
                    left_parent = NULL_REVISION
                tree.branch.set_last_revision_info(
                    len(tree.branch._lefthand_history(left_parent)),
                    left_parent)
                tree.commit(descendant, rev_id=descendant)
                pending.append(descendant)

    def test_lca(self):
        """Test finding least common ancestor.

        ancestry_1 should always have a single common ancestor
        """
        graph = self.make_graph(ancestry_1)
        self.assertRaises(errors.InvalidRevisionId, graph.find_lca, None)
        self.assertEqual({NULL_REVISION},
                         graph.find_lca(NULL_REVISION, NULL_REVISION))
        self.assertEqual({NULL_REVISION},
                         graph.find_lca(NULL_REVISION, b'rev1'))
        self.assertEqual({b'rev1'}, graph.find_lca(b'rev1', b'rev1'))
        self.assertEqual({b'rev1'}, graph.find_lca(b'rev2a', b'rev2b'))

    def test_no_unique_lca(self):
        """Test error when one revision is not in the graph"""
        graph = self.make_graph(ancestry_1)
        self.assertRaises(errors.NoCommonAncestor, graph.find_unique_lca,
                          b'rev1', b'1rev')

    def test_lca_criss_cross(self):
        """Test least-common-ancestor after a criss-cross merge."""
        graph = self.make_graph(criss_cross)
        self.assertEqual({b'rev2a', b'rev2b'},
                         graph.find_lca(b'rev3a', b'rev3b'))
        self.assertEqual({b'rev2b'},
                         graph.find_lca(b'rev3a', b'rev3b', b'rev2b'))

    def test_lca_shortcut(self):
        """Test least-common ancestor on this history shortcut"""
        graph = self.make_graph(history_shortcut)
        self.assertEqual({b'rev2b'}, graph.find_lca(b'rev3a', b'rev3b'))

    def test_lefthand_distance_smoke(self):
        """A simple does it work test for graph.lefthand_distance(keys)."""
        graph = self.make_graph(history_shortcut)
        distance_graph = graph.find_lefthand_distances([b'rev3b', b'rev2a'])
        self.assertEqual({b'rev2a': 2, b'rev3b': 3}, distance_graph)

    def test_lefthand_distance_ghosts(self):
        """A simple does it work test for graph.lefthand_distance(keys)."""
        nodes = {b'nonghost': [NULL_REVISION], b'toghost': [b'ghost']}
        graph = self.make_graph(nodes)
        distance_graph = graph.find_lefthand_distances(
            [b'nonghost', b'toghost'])
        self.assertEqual({b'nonghost': 1, b'toghost': -1}, distance_graph)

    def test_recursive_unique_lca(self):
        """Test finding a unique least common ancestor.

        ancestry_1 should always have a single common ancestor
        """
        graph = self.make_graph(ancestry_1)
        self.assertEqual(NULL_REVISION,
                         graph.find_unique_lca(NULL_REVISION, NULL_REVISION))
        self.assertEqual(NULL_REVISION,
                         graph.find_unique_lca(NULL_REVISION, b'rev1'))
        self.assertEqual(b'rev1', graph.find_unique_lca(b'rev1', b'rev1'))
        self.assertEqual(b'rev1', graph.find_unique_lca(b'rev2a', b'rev2b'))
        self.assertEqual((b'rev1', 1,),
                         graph.find_unique_lca(b'rev2a', b'rev2b',
                                               count_steps=True))

    def assertRemoveDescendants(self, expected, graph, revisions):
        parents = graph.get_parent_map(revisions)
        self.assertEqual(expected,
                         graph._remove_simple_descendants(revisions, parents))

    def test__remove_simple_descendants(self):
        graph = self.make_graph(ancestry_1)
        self.assertRemoveDescendants({b'rev1'}, graph,
                                     {b'rev1', b'rev2a', b'rev2b', b'rev3', b'rev4'})

    def test__remove_simple_descendants_disjoint(self):
        graph = self.make_graph(ancestry_1)
        self.assertRemoveDescendants({b'rev1', b'rev3'}, graph,
                                     {b'rev1', b'rev3'})

    def test__remove_simple_descendants_chain(self):
        graph = self.make_graph(ancestry_1)
        self.assertRemoveDescendants({b'rev1'}, graph,
                                     {b'rev1', b'rev2a', b'rev3'})

    def test__remove_simple_descendants_siblings(self):
        graph = self.make_graph(ancestry_1)
        self.assertRemoveDescendants({b'rev2a', b'rev2b'}, graph,
                                     {b'rev2a', b'rev2b', b'rev3'})

    def test_unique_lca_criss_cross(self):
        """Ensure we don't pick non-unique lcas in a criss-cross"""
        graph = self.make_graph(criss_cross)
        self.assertEqual(b'rev1', graph.find_unique_lca(b'rev3a', b'rev3b'))
        lca, steps = graph.find_unique_lca(
            b'rev3a', b'rev3b', count_steps=True)
        self.assertEqual(b'rev1', lca)
        self.assertEqual(2, steps)

    def test_unique_lca_null_revision(self):
        """Ensure we pick NULL_REVISION when necessary"""
        graph = self.make_graph(criss_cross2)
        self.assertEqual(b'rev1b', graph.find_unique_lca(b'rev2a', b'rev1b'))
        self.assertEqual(NULL_REVISION,
                         graph.find_unique_lca(b'rev2a', b'rev2b'))

    def test_unique_lca_null_revision2(self):
        """Ensure we pick NULL_REVISION when necessary"""
        graph = self.make_graph(ancestry_2)
        self.assertEqual(NULL_REVISION,
                         graph.find_unique_lca(b'rev4a', b'rev1b'))

    def test_lca_double_shortcut(self):
        graph = self.make_graph(double_shortcut)
        self.assertEqual(b'c', graph.find_unique_lca(b'f', b'g'))

    def test_common_ancestor_two_repos(self):
        """Ensure we do unique_lca using data from two repos"""
        mainline_tree = self.prepare_memory_tree('mainline')
        self.build_ancestry(mainline_tree, mainline)
        self.addCleanup(mainline_tree.unlock)

        # This is cheating, because the revisions in the graph are actually
        # different revisions, despite having the same revision-id.
        feature_tree = self.prepare_memory_tree('feature')
        self.build_ancestry(feature_tree, feature_branch)
        self.addCleanup(feature_tree.unlock)

        graph = mainline_tree.branch.repository.get_graph(
            feature_tree.branch.repository)
        self.assertEqual(b'rev2b', graph.find_unique_lca(b'rev2a', b'rev3b'))

    def test_graph_difference(self):
        graph = self.make_graph(ancestry_1)
        self.assertEqual(
            (set(), set()), graph.find_difference(b'rev1', b'rev1'))
        self.assertEqual((set(), {b'rev1'}),
                         graph.find_difference(NULL_REVISION, b'rev1'))
        self.assertEqual(({b'rev1'}, set()),
                         graph.find_difference(b'rev1', NULL_REVISION))
        self.assertEqual(({b'rev2a', b'rev3'}, {b'rev2b'}),
                         graph.find_difference(b'rev3', b'rev2b'))
        self.assertEqual(({b'rev4', b'rev3', b'rev2a'}, set()),
                         graph.find_difference(b'rev4', b'rev2b'))

    def test_graph_difference_separate_ancestry(self):
        graph = self.make_graph(ancestry_2)
        self.assertEqual(({b'rev1a'}, {b'rev1b'}),
                         graph.find_difference(b'rev1a', b'rev1b'))
        self.assertEqual(({b'rev1a', b'rev2a', b'rev3a', b'rev4a'},
                          {b'rev1b'}),
                         graph.find_difference(b'rev4a', b'rev1b'))

    def test_graph_difference_criss_cross(self):
        graph = self.make_graph(criss_cross)
        self.assertEqual(({b'rev3a'}, {b'rev3b'}),
                         graph.find_difference(b'rev3a', b'rev3b'))
        self.assertEqual((set(), {b'rev3b', b'rev2b'}),
                         graph.find_difference(b'rev2a', b'rev3b'))

    def test_graph_difference_extended_history(self):
        graph = self.make_graph(extended_history_shortcut)
        self.assertEqual(({b'e'}, {b'f'}),
                         graph.find_difference(b'e', b'f'))
        self.assertEqual(({b'f'}, {b'e'}),
                         graph.find_difference(b'f', b'e'))

    def test_graph_difference_double_shortcut(self):
        graph = self.make_graph(double_shortcut)
        self.assertEqual(({b'd', b'f'}, {b'e', b'g'}),
                         graph.find_difference(b'f', b'g'))

    def test_graph_difference_complex_shortcut(self):
        graph = self.make_graph(complex_shortcut)
        self.assertEqual(({b'm', b'i', b'e'}, {b'n', b'h'}),
                         graph.find_difference(b'm', b'n'))

    def test_graph_difference_complex_shortcut2(self):
        graph = self.make_graph(complex_shortcut2)
        self.assertEqual(({b't'}, {b'j', b'u'}),
                         graph.find_difference(b't', b'u'))

    def test_graph_difference_shortcut_extra_root(self):
        graph = self.make_graph(shortcut_extra_root)
        self.assertEqual(({b'e'}, {b'f', b'g'}),
                         graph.find_difference(b'e', b'f'))

    def test_iter_topo_order(self):
        graph = self.make_graph(ancestry_1)
        args = [b'rev2a', b'rev3', b'rev1']
        topo_args = list(graph.iter_topo_order(args))
        self.assertEqual(set(args), set(topo_args))
        self.assertTrue(topo_args.index(b'rev2a') > topo_args.index(b'rev1'))
        self.assertTrue(topo_args.index(b'rev2a') < topo_args.index(b'rev3'))

    def test_is_ancestor(self):
        graph = self.make_graph(ancestry_1)
        self.assertEqual(True, graph.is_ancestor(b'null:', b'null:'))
        self.assertEqual(True, graph.is_ancestor(b'null:', b'rev1'))
        self.assertEqual(False, graph.is_ancestor(b'rev1', b'null:'))
        self.assertEqual(True, graph.is_ancestor(b'null:', b'rev4'))
        self.assertEqual(False, graph.is_ancestor(b'rev4', b'null:'))
        self.assertEqual(False, graph.is_ancestor(b'rev4', b'rev2b'))
        self.assertEqual(True, graph.is_ancestor(b'rev2b', b'rev4'))
        self.assertEqual(False, graph.is_ancestor(b'rev2b', b'rev3'))
        self.assertEqual(False, graph.is_ancestor(b'rev3', b'rev2b'))
        instrumented_provider = InstrumentedParentsProvider(graph)
        instrumented_graph = _mod_graph.Graph(instrumented_provider)
        instrumented_graph.is_ancestor(b'rev2a', b'rev2b')
        self.assertTrue(b'null:' not in instrumented_provider.calls)

    def test_is_between(self):
        graph = self.make_graph(ancestry_1)
        self.assertEqual(True, graph.is_between(b'null:', b'null:', b'null:'))
        self.assertEqual(True, graph.is_between(b'rev1', b'null:', b'rev1'))
        self.assertEqual(True, graph.is_between(b'rev1', b'rev1', b'rev4'))
        self.assertEqual(True, graph.is_between(b'rev4', b'rev1', b'rev4'))
        self.assertEqual(True, graph.is_between(b'rev3', b'rev1', b'rev4'))
        self.assertEqual(False, graph.is_between(b'rev4', b'rev1', b'rev3'))
        self.assertEqual(False, graph.is_between(b'rev1', b'rev2a', b'rev4'))
        self.assertEqual(False, graph.is_between(b'null:', b'rev1', b'rev4'))

    def test_is_ancestor_boundary(self):
        """Ensure that we avoid searching the whole graph.

        This requires searching through b as a common ancestor, so we
        can identify that e is common.
        """
        graph = self.make_graph(boundary)
        instrumented_provider = InstrumentedParentsProvider(graph)
        graph = _mod_graph.Graph(instrumented_provider)
        self.assertFalse(graph.is_ancestor(b'a', b'c'))
        self.assertTrue(b'null:' not in instrumented_provider.calls)

    def test_iter_ancestry(self):
        nodes = boundary.copy()
        nodes[NULL_REVISION] = ()
        graph = self.make_graph(nodes)
        expected = nodes.copy()
        expected.pop(b'a')  # b'a' is not in the ancestry of b'c', all the
        # other nodes are
        self.assertEqual(expected, dict(graph.iter_ancestry([b'c'])))
        self.assertEqual(nodes, dict(graph.iter_ancestry([b'a', b'c'])))

    def test_iter_ancestry_with_ghost(self):
        graph = self.make_graph(with_ghost)
        expected = with_ghost.copy()
        # b'a' is not in the ancestry of b'c', and b'g' is a ghost
        expected[b'g'] = None
        self.assertEqual(expected, dict(graph.iter_ancestry([b'a', b'c'])))
        expected.pop(b'a')
        self.assertEqual(expected, dict(graph.iter_ancestry([b'c'])))

    def test_filter_candidate_lca(self):
        """Test filter_candidate_lca for a corner case

        This tests the case where we encounter the end of iteration for b'e'
        in the same pass as we discover that b'd' is an ancestor of b'e', and
        therefore b'e' can't be an lca.

        To compensate for different dict orderings on other Python
        implementations, we mirror b'd' and b'e' with b'b' and b'a'.
        """
        # This test is sensitive to the iteration order of dicts.  It will
        # pass incorrectly if b'e' and b'a' sort before b'c'
        #
        # NULL_REVISION
        #     / \
        #    a   e
        #    |   |
        #    b   d
        #     \ /
        #      c
        graph = self.make_graph({b'c': [b'b', b'd'], b'd': [b'e'], b'b': [b'a'],
                                 b'a': [NULL_REVISION], b'e': [NULL_REVISION]})
        self.assertEqual({b'c'}, graph.heads([b'a', b'c', b'e']))

    def test_heads_null(self):
        graph = self.make_graph(ancestry_1)
        self.assertEqual({b'null:'}, graph.heads([b'null:']))
        self.assertEqual({b'rev1'}, graph.heads([b'null:', b'rev1']))
        self.assertEqual({b'rev1'}, graph.heads([b'rev1', b'null:']))
        self.assertEqual({b'rev1'}, graph.heads({b'rev1', b'null:'}))
        self.assertEqual({b'rev1'}, graph.heads((b'rev1', b'null:')))

    def test_heads_one(self):
        # A single node will always be a head
        graph = self.make_graph(ancestry_1)
        self.assertEqual({b'null:'}, graph.heads([b'null:']))
        self.assertEqual({b'rev1'}, graph.heads([b'rev1']))
        self.assertEqual({b'rev2a'}, graph.heads([b'rev2a']))
        self.assertEqual({b'rev2b'}, graph.heads([b'rev2b']))
        self.assertEqual({b'rev3'}, graph.heads([b'rev3']))
        self.assertEqual({b'rev4'}, graph.heads([b'rev4']))

    def test_heads_single(self):
        graph = self.make_graph(ancestry_1)
        self.assertEqual({b'rev4'}, graph.heads([b'null:', b'rev4']))
        self.assertEqual({b'rev2a'}, graph.heads([b'rev1', b'rev2a']))
        self.assertEqual({b'rev2b'}, graph.heads([b'rev1', b'rev2b']))
        self.assertEqual({b'rev3'}, graph.heads([b'rev1', b'rev3']))
        self.assertEqual({b'rev4'}, graph.heads([b'rev1', b'rev4']))
        self.assertEqual({b'rev4'}, graph.heads([b'rev2a', b'rev4']))
        self.assertEqual({b'rev4'}, graph.heads([b'rev2b', b'rev4']))
        self.assertEqual({b'rev4'}, graph.heads([b'rev3', b'rev4']))

    def test_heads_two_heads(self):
        graph = self.make_graph(ancestry_1)
        self.assertEqual({b'rev2a', b'rev2b'},
                         graph.heads([b'rev2a', b'rev2b']))
        self.assertEqual({b'rev3', b'rev2b'},
                         graph.heads([b'rev3', b'rev2b']))

    def test_heads_criss_cross(self):
        graph = self.make_graph(criss_cross)
        self.assertEqual({b'rev2a'},
                         graph.heads([b'rev2a', b'rev1']))
        self.assertEqual({b'rev2b'},
                         graph.heads([b'rev2b', b'rev1']))
        self.assertEqual({b'rev3a'},
                         graph.heads([b'rev3a', b'rev1']))
        self.assertEqual({b'rev3b'},
                         graph.heads([b'rev3b', b'rev1']))
        self.assertEqual({b'rev2a', b'rev2b'},
                         graph.heads([b'rev2a', b'rev2b']))
        self.assertEqual({b'rev3a'},
                         graph.heads([b'rev3a', b'rev2a']))
        self.assertEqual({b'rev3a'},
                         graph.heads([b'rev3a', b'rev2b']))
        self.assertEqual({b'rev3a'},
                         graph.heads([b'rev3a', b'rev2a', b'rev2b']))
        self.assertEqual({b'rev3b'},
                         graph.heads([b'rev3b', b'rev2a']))
        self.assertEqual({b'rev3b'},
                         graph.heads([b'rev3b', b'rev2b']))
        self.assertEqual({b'rev3b'},
                         graph.heads([b'rev3b', b'rev2a', b'rev2b']))
        self.assertEqual({b'rev3a', b'rev3b'},
                         graph.heads([b'rev3a', b'rev3b']))
        self.assertEqual({b'rev3a', b'rev3b'},
                         graph.heads([b'rev3a', b'rev3b', b'rev2a', b'rev2b']))

    def test_heads_shortcut(self):
        graph = self.make_graph(history_shortcut)

        self.assertEqual({b'rev2a', b'rev2b', b'rev2c'},
                         graph.heads([b'rev2a', b'rev2b', b'rev2c']))
        self.assertEqual({b'rev3a', b'rev3b'},
                         graph.heads([b'rev3a', b'rev3b']))
        self.assertEqual({b'rev3a', b'rev3b'},
                         graph.heads([b'rev2a', b'rev3a', b'rev3b']))
        self.assertEqual({b'rev2a', b'rev3b'},
                         graph.heads([b'rev2a', b'rev3b']))
        self.assertEqual({b'rev2c', b'rev3a'},
                         graph.heads([b'rev2c', b'rev3a']))

    def _run_heads_break_deeper(self, graph_dict, search):
        """Run heads on a graph-as-a-dict.

        If the search asks for the parents of b'deeper' the test will fail.
        """
        class stub:
            pass

        def get_parent_map(keys):
            result = {}
            for key in keys:
                if key == b'deeper':
                    self.fail('key deeper was accessed')
                result[key] = graph_dict[key]
            return result
        an_obj = stub()
        an_obj.get_parent_map = get_parent_map
        graph = _mod_graph.Graph(an_obj)
        return graph.heads(search)

    def test_heads_limits_search(self):
        # test that a heads query does not search all of history
        graph_dict = {
            b'left': [b'common'],
            b'right': [b'common'],
            b'common': [b'deeper'],
        }
        self.assertEqual({b'left', b'right'},
                         self._run_heads_break_deeper(graph_dict, [b'left', b'right']))

    def test_heads_limits_search_assymetric(self):
        # test that a heads query does not search all of history
        graph_dict = {
            b'left': [b'midleft'],
            b'midleft': [b'common'],
            b'right': [b'common'],
            b'common': [b'aftercommon'],
            b'aftercommon': [b'deeper'],
        }
        self.assertEqual({b'left', b'right'},
                         self._run_heads_break_deeper(graph_dict, [b'left', b'right']))

    def test_heads_limits_search_common_search_must_continue(self):
        # test that common nodes are still queried, preventing
        # all-the-way-to-origin behaviour in the following graph:
        graph_dict = {
            b'h1': [b'shortcut', b'common1'],
            b'h2': [b'common1'],
            b'shortcut': [b'common2'],
            b'common1': [b'common2'],
            b'common2': [b'deeper'],
        }
        self.assertEqual({b'h1', b'h2'},
                         self._run_heads_break_deeper(graph_dict, [b'h1', b'h2']))

    def test_breadth_first_search_start_ghosts(self):
        graph = self.make_graph({})
        # with_ghosts reports the ghosts
        search = graph._make_breadth_first_searcher([b'a-ghost'])
        self.assertEqual((set(), {b'a-ghost'}), search.next_with_ghosts())
        self.assertRaises(StopIteration, search.next_with_ghosts)
        # next includes them
        search = graph._make_breadth_first_searcher([b'a-ghost'])
        self.assertEqual({b'a-ghost'}, next(search))
        self.assertRaises(StopIteration, next, search)

    def test_breadth_first_search_deep_ghosts(self):
        graph = self.make_graph({
            b'head': [b'present'],
            b'present': [b'child', b'ghost'],
            b'child': [],
            })
        # with_ghosts reports the ghosts
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertEqual(({b'head'}, set()), search.next_with_ghosts())
        self.assertEqual(({b'present'}, set()), search.next_with_ghosts())
        self.assertEqual(({b'child'}, {b'ghost'}),
                         search.next_with_ghosts())
        self.assertRaises(StopIteration, search.next_with_ghosts)
        # next includes them
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertEqual({b'head'}, next(search))
        self.assertEqual({b'present'}, next(search))
        self.assertEqual({b'child', b'ghost'},
                         next(search))
        self.assertRaises(StopIteration, next, search)

    def test_breadth_first_search_change_next_to_next_with_ghosts(self):
        # To make the API robust, we allow calling both next() and
        # next_with_ghosts() on the same searcher.
        graph = self.make_graph({
            b'head': [b'present'],
            b'present': [b'child', b'ghost'],
            b'child': [],
            })
        # start with next_with_ghosts
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertEqual(({b'head'}, set()), search.next_with_ghosts())
        self.assertEqual({b'present'}, next(search))
        self.assertEqual(({b'child'}, {b'ghost'}),
                         search.next_with_ghosts())
        self.assertRaises(StopIteration, next, search)
        # start with next
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertEqual({b'head'}, next(search))
        self.assertEqual(({b'present'}, set()), search.next_with_ghosts())
        self.assertEqual({b'child', b'ghost'},
                         next(search))
        self.assertRaises(StopIteration, search.next_with_ghosts)

    def test_breadth_first_change_search(self):
        # Changing the search should work with both next and next_with_ghosts.
        graph = self.make_graph({
            b'head': [b'present'],
            b'present': [b'stopped'],
            b'other': [b'other_2'],
            b'other_2': [],
            })
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertEqual(({b'head'}, set()), search.next_with_ghosts())
        self.assertEqual(({b'present'}, set()), search.next_with_ghosts())
        self.assertEqual({b'present'},
                         search.stop_searching_any([b'present']))
        self.assertEqual(({b'other'}, {b'other_ghost'}),
                         search.start_searching([b'other', b'other_ghost']))
        self.assertEqual(({b'other_2'}, set()), search.next_with_ghosts())
        self.assertRaises(StopIteration, search.next_with_ghosts)
        # next includes them
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertEqual({b'head'}, next(search))
        self.assertEqual({b'present'}, next(search))
        self.assertEqual({b'present'},
                         search.stop_searching_any([b'present']))
        search.start_searching([b'other', b'other_ghost'])
        self.assertEqual({b'other_2'}, next(search))
        self.assertRaises(StopIteration, next, search)

    def assertSeenAndResult(self, instructions, search, next):
        """Check the results of .seen and get_result() for a seach.

        :param instructions: A list of tuples:
            (seen, recipe, included_keys, starts, stops).
            seen, recipe and included_keys are results to check on the search
            and the searches get_result(). starts and stops are parameters to
            pass to start_searching and stop_searching_any during each
            iteration, if they are not None.
        :param search: The search to use.
        :param next: A callable to advance the search.
        """
        for seen, recipe, included_keys, starts, stops in instructions:
            # Adjust for recipe contract changes that don't vary for all the
            # current tests.
            recipe = ('search',) + recipe
            next()
            if starts is not None:
                search.start_searching(starts)
            if stops is not None:
                search.stop_searching_any(stops)
            state = search.get_state()
            self.assertEqual(set(included_keys), state[2])
            self.assertEqual(seen, search.seen)

    def test_breadth_first_get_result_excludes_current_pending(self):
        graph = self.make_graph({
            b'head': [b'child'],
            b'child': [NULL_REVISION],
            NULL_REVISION: [],
            })
        search = graph._make_breadth_first_searcher([b'head'])
        # At the start, nothing has been seen, to its all excluded:
        state = search.get_state()
        self.assertEqual(({b'head'}, {b'head'}, set()),
                         state)
        self.assertEqual(set(), search.seen)
        # using next:
        expected = [
            ({b'head'}, ({b'head'}, {b'child'}, 1),
             [b'head'], None, None),
            ({b'head', b'child'}, ({b'head'}, {NULL_REVISION}, 2),
             [b'head', b'child'], None, None),
            ({b'head', b'child', NULL_REVISION}, ({b'head'}, set(), 3),
             [b'head', b'child', NULL_REVISION], None, None),
            ]
        self.assertSeenAndResult(expected, search, search.__next__)
        # using next_with_ghosts:
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertSeenAndResult(expected, search, search.next_with_ghosts)

    def test_breadth_first_get_result_starts_stops(self):
        graph = self.make_graph({
            b'head': [b'child'],
            b'child': [NULL_REVISION],
            b'otherhead': [b'otherchild'],
            b'otherchild': [b'excluded'],
            b'excluded': [NULL_REVISION],
            NULL_REVISION: []
            })
        search = graph._make_breadth_first_searcher([])
        # Starting with nothing and adding a search works:
        search.start_searching([b'head'])
        # head has been seen:
        state = search.get_state()
        self.assertEqual(({b'head'}, {b'child'}, {b'head'}),
                         state)
        self.assertEqual({b'head'}, search.seen)
        # using next:
        expected = [
            # stop at child, and start a new search at otherhead:
            # - otherhead counts as seen immediately when start_searching is
            # called.
            ({b'head', b'child', b'otherhead'},
             ({b'head', b'otherhead'}, {b'child', b'otherchild'}, 2),
             [b'head', b'otherhead'], [b'otherhead'], [b'child']),
            ({b'head', b'child', b'otherhead', b'otherchild'},
             ({b'head', b'otherhead'}, {b'child', b'excluded'}, 3),
             [b'head', b'otherhead', b'otherchild'], None, None),
            # stop searching excluded now
            ({b'head', b'child', b'otherhead', b'otherchild', b'excluded'},
             ({b'head', b'otherhead'}, {b'child', b'excluded'}, 3),
             [b'head', b'otherhead', b'otherchild'], None, [b'excluded']),
            ]
        self.assertSeenAndResult(expected, search, search.__next__)
        # using next_with_ghosts:
        search = graph._make_breadth_first_searcher([])
        search.start_searching([b'head'])
        self.assertSeenAndResult(expected, search, search.next_with_ghosts)

    def test_breadth_first_stop_searching_not_queried(self):
        # A client should be able to say b'stop node X' even if X has not been
        # returned to the client.
        graph = self.make_graph({
            b'head': [b'child', b'ghost1'],
            b'child': [NULL_REVISION],
            NULL_REVISION: [],
            })
        search = graph._make_breadth_first_searcher([b'head'])
        expected = [
            # NULL_REVISION and ghost1 have not been returned
            ({b'head'},
             ({b'head'}, {b'child', NULL_REVISION, b'ghost1'}, 1),
             [b'head'], None, [NULL_REVISION, b'ghost1']),
            # ghost1 has been returned, NULL_REVISION is to be returned in the
            # next iteration.
            ({b'head', b'child', b'ghost1'},
             ({b'head'}, {b'ghost1', NULL_REVISION}, 2),
             [b'head', b'child'], None, [NULL_REVISION, b'ghost1']),
            ]
        self.assertSeenAndResult(expected, search, search.__next__)
        # using next_with_ghosts:
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertSeenAndResult(expected, search, search.next_with_ghosts)

    def test_breadth_first_stop_searching_late(self):
        # A client should be able to say b'stop node X' and have it excluded
        # from the result even if X was seen in an older iteration of the
        # search.
        graph = self.make_graph({
            b'head': [b'middle'],
            b'middle': [b'child'],
            b'child': [NULL_REVISION],
            NULL_REVISION: [],
            })
        search = graph._make_breadth_first_searcher([b'head'])
        expected = [
            ({b'head'}, ({b'head'}, {b'middle'}, 1),
             [b'head'], None, None),
            ({b'head', b'middle'}, ({b'head'}, {b'child'}, 2),
             [b'head', b'middle'], None, None),
            # b'middle' came from the previous iteration, but we don't stop
            # searching it until *after* advancing the searcher.
            ({b'head', b'middle', b'child'},
             ({b'head'}, {b'middle', b'child'}, 1),
             [b'head'], None, [b'middle', b'child']),
            ]
        self.assertSeenAndResult(expected, search, search.__next__)
        # using next_with_ghosts:
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertSeenAndResult(expected, search, search.next_with_ghosts)

    def test_breadth_first_get_result_ghosts_are_excluded(self):
        graph = self.make_graph({
            b'head': [b'child', b'ghost'],
            b'child': [NULL_REVISION],
            NULL_REVISION: [],
            })
        search = graph._make_breadth_first_searcher([b'head'])
        # using next:
        expected = [
            ({b'head'},
             ({b'head'}, {b'ghost', b'child'}, 1),
             [b'head'], None, None),
            ({b'head', b'child', b'ghost'},
             ({b'head'}, {NULL_REVISION, b'ghost'}, 2),
             [b'head', b'child'], None, None),
            ]
        self.assertSeenAndResult(expected, search, search.__next__)
        # using next_with_ghosts:
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertSeenAndResult(expected, search, search.next_with_ghosts)

    def test_breadth_first_get_result_starting_a_ghost_ghost_is_excluded(self):
        graph = self.make_graph({
            b'head': [b'child'],
            b'child': [NULL_REVISION],
            NULL_REVISION: [],
            })
        search = graph._make_breadth_first_searcher([b'head'])
        # using next:
        expected = [
            ({b'head', b'ghost'},
             ({b'head', b'ghost'}, {b'child', b'ghost'}, 1),
             [b'head'], [b'ghost'], None),
            ({b'head', b'child', b'ghost'},
             ({b'head', b'ghost'}, {NULL_REVISION, b'ghost'}, 2),
             [b'head', b'child'], None, None),
            ]
        self.assertSeenAndResult(expected, search, search.__next__)
        # using next_with_ghosts:
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertSeenAndResult(expected, search, search.next_with_ghosts)

    def test_breadth_first_revision_count_includes_NULL_REVISION(self):
        graph = self.make_graph({
            b'head': [NULL_REVISION],
            NULL_REVISION: [],
            })
        search = graph._make_breadth_first_searcher([b'head'])
        # using next:
        expected = [
            ({b'head'},
             ({b'head'}, {NULL_REVISION}, 1),
             [b'head'], None, None),
            ({b'head', NULL_REVISION},
             ({b'head'}, set(), 2),
             [b'head', NULL_REVISION], None, None),
            ]
        self.assertSeenAndResult(expected, search, search.__next__)
        # using next_with_ghosts:
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertSeenAndResult(expected, search, search.next_with_ghosts)

    def test_breadth_first_search_get_result_after_StopIteration(self):
        # StopIteration should not invalid anything..
        graph = self.make_graph({
            b'head': [NULL_REVISION],
            NULL_REVISION: [],
            })
        search = graph._make_breadth_first_searcher([b'head'])
        # using next:
        expected = [
            ({b'head'},
             ({b'head'}, {NULL_REVISION}, 1),
             [b'head'], None, None),
            ({b'head', b'ghost', NULL_REVISION},
             ({b'head', b'ghost'}, {b'ghost'}, 2),
             [b'head', NULL_REVISION], [b'ghost'], None),
            ]
        self.assertSeenAndResult(expected, search, search.__next__)
        self.assertRaises(StopIteration, next, search)
        self.assertEqual({b'head', b'ghost', NULL_REVISION}, search.seen)
        state = search.get_state()
        self.assertEqual(
            ({b'ghost', b'head'}, {b'ghost'},
                {b'head', NULL_REVISION}),
            state)
        # using next_with_ghosts:
        search = graph._make_breadth_first_searcher([b'head'])
        self.assertSeenAndResult(expected, search, search.next_with_ghosts)
        self.assertRaises(StopIteration, next, search)
        self.assertEqual({b'head', b'ghost', NULL_REVISION}, search.seen)
        state = search.get_state()
        self.assertEqual(
            ({b'ghost', b'head'}, {b'ghost'},
                {b'head', NULL_REVISION}),
            state)


class TestFindUniqueAncestors(TestGraphBase):

    def assertFindUniqueAncestors(self, graph, expected, node, common):
        actual = graph.find_unique_ancestors(node, common)
        self.assertEqual(expected, sorted(actual))

    def test_empty_set(self):
        graph = self.make_graph(ancestry_1)
        self.assertFindUniqueAncestors(graph, [], b'rev1', [b'rev1'])
        self.assertFindUniqueAncestors(graph, [], b'rev2b', [b'rev2b'])
        self.assertFindUniqueAncestors(graph, [], b'rev3', [b'rev1', b'rev3'])

    def test_single_node(self):
        graph = self.make_graph(ancestry_1)
        self.assertFindUniqueAncestors(graph, [b'rev2a'], b'rev2a', [b'rev1'])
        self.assertFindUniqueAncestors(graph, [b'rev2b'], b'rev2b', [b'rev1'])
        self.assertFindUniqueAncestors(graph, [b'rev3'], b'rev3', [b'rev2a'])

    def test_minimal_ancestry(self):
        graph = self.make_breaking_graph(extended_history_shortcut,
                                         [NULL_REVISION, b'a', b'b'])
        self.assertFindUniqueAncestors(graph, [b'e'], b'e', [b'd'])

        graph = self.make_breaking_graph(extended_history_shortcut,
                                         [b'b'])
        self.assertFindUniqueAncestors(graph, [b'f'], b'f', [b'a', b'd'])

        graph = self.make_breaking_graph(complex_shortcut,
                                         [b'a', b'b'])
        self.assertFindUniqueAncestors(graph, [b'h'], b'h', [b'i'])
        self.assertFindUniqueAncestors(graph, [b'e', b'g', b'i'], b'i', [b'h'])
        self.assertFindUniqueAncestors(graph, [b'h'], b'h', [b'g'])
        self.assertFindUniqueAncestors(graph, [b'h'], b'h', [b'j'])

    def test_in_ancestry(self):
        graph = self.make_graph(ancestry_1)
        self.assertFindUniqueAncestors(graph, [], b'rev1', [b'rev3'])
        self.assertFindUniqueAncestors(graph, [], b'rev2b', [b'rev4'])

    def test_multiple_revisions(self):
        graph = self.make_graph(ancestry_1)
        self.assertFindUniqueAncestors(graph,
                                       [b'rev4'], b'rev4', [b'rev3', b'rev2b'])
        self.assertFindUniqueAncestors(graph,
                                       [b'rev2a', b'rev3', b'rev4'], b'rev4', [b'rev2b'])

    def test_complex_shortcut(self):
        graph = self.make_graph(complex_shortcut)
        self.assertFindUniqueAncestors(graph,
                                       [b'h', b'n'], b'n', [b'm'])
        self.assertFindUniqueAncestors(graph,
                                       [b'e', b'i', b'm'], b'm', [b'n'])

    def test_complex_shortcut2(self):
        graph = self.make_graph(complex_shortcut2)
        self.assertFindUniqueAncestors(graph,
                                       [b'j', b'u'], b'u', [b't'])
        self.assertFindUniqueAncestors(graph,
                                       [b't'], b't', [b'u'])

    def test_multiple_interesting_unique(self):
        graph = self.make_graph(multiple_interesting_unique)
        self.assertFindUniqueAncestors(graph,
                                       [b'j', b'y'], b'y', [b'z'])
        self.assertFindUniqueAncestors(graph,
                                       [b'p', b'z'], b'z', [b'y'])

    def test_racing_shortcuts(self):
        graph = self.make_graph(racing_shortcuts)
        self.assertFindUniqueAncestors(graph,
                                       [b'p', b'q', b'z'], b'z', [b'y'])
        self.assertFindUniqueAncestors(graph,
                                       [b'h', b'i', b'j', b'y'], b'j', [b'z'])


class TestGraphFindDistanceToNull(TestGraphBase):
    """Test an api that should be able to compute a revno"""

    def assertFindDistance(self, revno, graph, target_id, known_ids):
        """Assert the output of Graph.find_distance_to_null()"""
        actual = graph.find_distance_to_null(target_id, known_ids)
        self.assertEqual(revno, actual)

    def test_nothing_known(self):
        graph = self.make_graph(ancestry_1)
        self.assertFindDistance(0, graph, NULL_REVISION, [])
        self.assertFindDistance(1, graph, b'rev1', [])
        self.assertFindDistance(2, graph, b'rev2a', [])
        self.assertFindDistance(2, graph, b'rev2b', [])
        self.assertFindDistance(3, graph, b'rev3', [])
        self.assertFindDistance(4, graph, b'rev4', [])

    def test_rev_is_ghost(self):
        graph = self.make_graph(ancestry_1)
        e = self.assertRaises(errors.GhostRevisionsHaveNoRevno,
                              graph.find_distance_to_null, b'rev_missing', [])
        self.assertEqual(b'rev_missing', e.revision_id)
        self.assertEqual(b'rev_missing', e.ghost_revision_id)

    def test_ancestor_is_ghost(self):
        graph = self.make_graph({b'rev': [b'parent']})
        e = self.assertRaises(errors.GhostRevisionsHaveNoRevno,
                              graph.find_distance_to_null, b'rev', [])
        self.assertEqual(b'rev', e.revision_id)
        self.assertEqual(b'parent', e.ghost_revision_id)

    def test_known_in_ancestry(self):
        graph = self.make_graph(ancestry_1)
        self.assertFindDistance(2, graph, b'rev2a', [(b'rev1', 1)])
        self.assertFindDistance(3, graph, b'rev3', [(b'rev2a', 2)])

    def test_known_in_ancestry_limits(self):
        graph = self.make_breaking_graph(ancestry_1, [b'rev1'])
        self.assertFindDistance(4, graph, b'rev4', [(b'rev3', 3)])

    def test_target_is_ancestor(self):
        graph = self.make_graph(ancestry_1)
        self.assertFindDistance(2, graph, b'rev2a', [(b'rev3', 3)])

    def test_target_is_ancestor_limits(self):
        """We shouldn't search all history if we run into ourselves"""
        graph = self.make_breaking_graph(ancestry_1, [b'rev1'])
        self.assertFindDistance(3, graph, b'rev3', [(b'rev4', 4)])

    def test_target_parallel_to_known_limits(self):
        # Even though the known revision isn't part of the other ancestry, they
        # eventually converge
        graph = self.make_breaking_graph(with_tail, [b'a'])
        self.assertFindDistance(6, graph, b'f', [(b'g', 6)])
        self.assertFindDistance(7, graph, b'h', [(b'g', 6)])
        self.assertFindDistance(8, graph, b'i', [(b'g', 6)])
        self.assertFindDistance(6, graph, b'g', [(b'i', 8)])


class TestFindMergeOrder(TestGraphBase):

    def assertMergeOrder(self, expected, graph, tip, base_revisions):
        self.assertEqual(expected, graph.find_merge_order(tip, base_revisions))

    def test_parents(self):
        graph = self.make_graph(ancestry_1)
        self.assertMergeOrder([b'rev3', b'rev2b'], graph, b'rev4',
                              [b'rev3', b'rev2b'])
        self.assertMergeOrder([b'rev3', b'rev2b'], graph, b'rev4',
                              [b'rev2b', b'rev3'])

    def test_ancestors(self):
        graph = self.make_graph(ancestry_1)
        self.assertMergeOrder([b'rev1', b'rev2b'], graph, b'rev4',
                              [b'rev1', b'rev2b'])
        self.assertMergeOrder([b'rev1', b'rev2b'], graph, b'rev4',
                              [b'rev2b', b'rev1'])

    def test_shortcut_one_ancestor(self):
        # When we have enough info, we can stop searching
        graph = self.make_breaking_graph(
            ancestry_1, [b'rev3', b'rev2b', b'rev4'])
        # Single ancestors shortcut right away
        self.assertMergeOrder([b'rev3'], graph, b'rev4', [b'rev3'])

    def test_shortcut_after_one_ancestor(self):
        graph = self.make_breaking_graph(ancestry_1, [b'rev2a', b'rev2b'])
        self.assertMergeOrder([b'rev3', b'rev1'], graph,
                              b'rev4', [b'rev1', b'rev3'])


class TestFindDescendants(TestGraphBase):

    def test_find_descendants_rev1_rev3(self):
        graph = self.make_graph(ancestry_1)
        descendants = graph.find_descendants(b'rev1', b'rev3')
        self.assertEqual({b'rev1', b'rev2a', b'rev3'}, descendants)

    def test_find_descendants_rev1_rev4(self):
        graph = self.make_graph(ancestry_1)
        descendants = graph.find_descendants(b'rev1', b'rev4')
        self.assertEqual({b'rev1', b'rev2a', b'rev2b', b'rev3', b'rev4'},
                         descendants)

    def test_find_descendants_rev2a_rev4(self):
        graph = self.make_graph(ancestry_1)
        descendants = graph.find_descendants(b'rev2a', b'rev4')
        self.assertEqual({b'rev2a', b'rev3', b'rev4'}, descendants)


class TestFindLefthandMerger(TestGraphBase):

    def check_merger(self, result, ancestry, merged, tip):
        graph = self.make_graph(ancestry)
        self.assertEqual(result, graph.find_lefthand_merger(merged, tip))

    def test_find_lefthand_merger_rev2b(self):
        self.check_merger(b'rev4', ancestry_1, b'rev2b', b'rev4')

    def test_find_lefthand_merger_rev2a(self):
        self.check_merger(b'rev2a', ancestry_1, b'rev2a', b'rev4')

    def test_find_lefthand_merger_rev4(self):
        self.check_merger(None, ancestry_1, b'rev4', b'rev2a')

    def test_find_lefthand_merger_f(self):
        self.check_merger(b'i', complex_shortcut, b'f', b'm')

    def test_find_lefthand_merger_g(self):
        self.check_merger(b'i', complex_shortcut, b'g', b'm')

    def test_find_lefthand_merger_h(self):
        self.check_merger(b'n', complex_shortcut, b'h', b'n')


class TestGetChildMap(TestGraphBase):

    def test_get_child_map(self):
        graph = self.make_graph(ancestry_1)
        child_map = graph.get_child_map([b'rev4', b'rev3', b'rev2a', b'rev2b'])
        self.assertEqual({b'rev1': [b'rev2a', b'rev2b'],
                          b'rev2a': [b'rev3'],
                          b'rev2b': [b'rev4'],
                          b'rev3': [b'rev4']},
                         child_map)


class TestCachingParentsProvider(tests.TestCase):
    """These tests run with:

    self.inst_pp, a recording parents provider with a graph of a->b, and b is a
    ghost.
    self.caching_pp, a CachingParentsProvider layered on inst_pp.
    """

    def setUp(self):
        super().setUp()
        dict_pp = _mod_graph.DictParentsProvider({b'a': (b'b',)})
        self.inst_pp = InstrumentedParentsProvider(dict_pp)
        self.caching_pp = _mod_graph.CachingParentsProvider(self.inst_pp)

    def test_get_parent_map(self):
        """Requesting the same revision should be returned from cache"""
        self.assertEqual({}, self.caching_pp._cache)
        self.assertEqual(
            {b'a': (b'b',)}, self.caching_pp.get_parent_map([b'a']))
        self.assertEqual([b'a'], self.inst_pp.calls)
        self.assertEqual(
            {b'a': (b'b',)}, self.caching_pp.get_parent_map([b'a']))
        # No new call, as it should have been returned from the cache
        self.assertEqual([b'a'], self.inst_pp.calls)
        self.assertEqual({b'a': (b'b',)}, self.caching_pp._cache)

    def test_get_parent_map_not_present(self):
        """The cache should also track when a revision doesn't exist"""
        self.assertEqual({}, self.caching_pp.get_parent_map([b'b']))
        self.assertEqual([b'b'], self.inst_pp.calls)
        self.assertEqual({}, self.caching_pp.get_parent_map([b'b']))
        # No new calls
        self.assertEqual([b'b'], self.inst_pp.calls)

    def test_get_parent_map_mixed(self):
        """Anything that can be returned from cache, should be"""
        self.assertEqual({}, self.caching_pp.get_parent_map([b'b']))
        self.assertEqual([b'b'], self.inst_pp.calls)
        self.assertEqual({b'a': (b'b',)},
                         self.caching_pp.get_parent_map([b'a', b'b']))
        self.assertEqual([b'b', b'a'], self.inst_pp.calls)

    def test_get_parent_map_repeated(self):
        """Asking for the same parent 2x will only forward 1 request."""
        self.assertEqual({b'a': (b'b',)},
                         self.caching_pp.get_parent_map([b'b', b'a', b'b']))
        # Use sorted because we don't care about the order, just that each is
        # only present 1 time.
        self.assertEqual([b'a', b'b'], sorted(self.inst_pp.calls))

    def test_note_missing_key(self):
        """After noting that a key is missing it is cached."""
        self.caching_pp.note_missing_key(b'b')
        self.assertEqual({}, self.caching_pp.get_parent_map([b'b']))
        self.assertEqual([], self.inst_pp.calls)
        self.assertEqual({b'b'}, self.caching_pp.missing_keys)

    def test_get_cached_parent_map(self):
        self.assertEqual({}, self.caching_pp.get_cached_parent_map([b'a']))
        self.assertEqual([], self.inst_pp.calls)
        self.assertEqual(
            {b'a': (b'b',)}, self.caching_pp.get_parent_map([b'a']))
        self.assertEqual([b'a'], self.inst_pp.calls)
        self.assertEqual({b'a': (b'b',)},
                         self.caching_pp.get_cached_parent_map([b'a']))


class TestCachingParentsProviderExtras(tests.TestCaseWithTransport):
    """Test the behaviour when parents are provided that were not requested."""

    def setUp(self):
        super().setUp()

        class ExtraParentsProvider:

            def get_parent_map(self, keys):
                return {b'rev1': [], b'rev2': [b'rev1', ]}

        self.inst_pp = InstrumentedParentsProvider(ExtraParentsProvider())
        self.caching_pp = _mod_graph.CachingParentsProvider(
            get_parent_map=self.inst_pp.get_parent_map)

    def test_uncached(self):
        self.caching_pp.disable_cache()
        self.assertEqual({b'rev1': []},
                         self.caching_pp.get_parent_map([b'rev1']))
        self.assertEqual([b'rev1'], self.inst_pp.calls)
        self.assertIs(None, self.caching_pp._cache)

    def test_cache_initially_empty(self):
        self.assertEqual({}, self.caching_pp._cache)

    def test_cached(self):
        self.assertEqual({b'rev1': []},
                         self.caching_pp.get_parent_map([b'rev1']))
        self.assertEqual([b'rev1'], self.inst_pp.calls)
        self.assertEqual({b'rev1': [], b'rev2': [b'rev1']},
                         self.caching_pp._cache)
        self.assertEqual({b'rev1': []},
                         self.caching_pp.get_parent_map([b'rev1']))
        self.assertEqual([b'rev1'], self.inst_pp.calls)

    def test_disable_cache_clears_cache(self):
        # Put something in the cache
        self.caching_pp.get_parent_map([b'rev1'])
        self.assertEqual(2, len(self.caching_pp._cache))
        self.caching_pp.disable_cache()
        self.assertIs(None, self.caching_pp._cache)

    def test_enable_cache_raises(self):
        e = self.assertRaises(AssertionError, self.caching_pp.enable_cache)
        self.assertEqual('Cache enabled when already enabled.', str(e))

    def test_cache_misses(self):
        self.caching_pp.get_parent_map([b'rev3'])
        self.caching_pp.get_parent_map([b'rev3'])
        self.assertEqual([b'rev3'], self.inst_pp.calls)

    def test_no_cache_misses(self):
        self.caching_pp.disable_cache()
        self.caching_pp.enable_cache(cache_misses=False)
        self.caching_pp.get_parent_map([b'rev3'])
        self.caching_pp.get_parent_map([b'rev3'])
        self.assertEqual([b'rev3', b'rev3'], self.inst_pp.calls)

    def test_cache_extras(self):
        self.assertEqual({}, self.caching_pp.get_parent_map([b'rev3']))
        self.assertEqual({b'rev2': [b'rev1']},
                         self.caching_pp.get_parent_map([b'rev2']))
        self.assertEqual([b'rev3'], self.inst_pp.calls)

    def test_extras_using_cached(self):
        self.assertEqual({}, self.caching_pp.get_cached_parent_map([b'rev3']))
        self.assertEqual({}, self.caching_pp.get_parent_map([b'rev3']))
        self.assertEqual({b'rev2': [b'rev1']},
                         self.caching_pp.get_cached_parent_map([b'rev2']))
        self.assertEqual([b'rev3'], self.inst_pp.calls)


class TestCollapseLinearRegions(tests.TestCase):

    def assertCollapsed(self, collapsed, original):
        self.assertEqual(collapsed,
                         _mod_graph.collapse_linear_regions(original))

    def test_collapse_nothing(self):
        d = {1: [2, 3], 2: [], 3: []}
        self.assertCollapsed(d, d)
        d = {1: [2], 2: [3, 4], 3: [5], 4: [5], 5: []}
        self.assertCollapsed(d, d)

    def test_collapse_chain(self):
        # Any time we have a linear chain, we should be able to collapse
        d = {1: [2], 2: [3], 3: [4], 4: [5], 5: []}
        self.assertCollapsed({1: [5], 5: []}, d)
        d = {5: [4], 4: [3], 3: [2], 2: [1], 1: []}
        self.assertCollapsed({5: [1], 1: []}, d)
        d = {5: [3], 3: [4], 4: [1], 1: [2], 2: []}
        self.assertCollapsed({5: [2], 2: []}, d)

    def test_collapse_with_multiple_children(self):
        #    7
        #    |
        #    6
        #   / \
        #  4   5
        #  |   |
        #  2   3
        #   \ /
        #    1
        #
        # 4 and 5 cannot be removed because 6 has 2 children
        # 2 and 3 cannot be removed because 1 has 2 parents
        d = {1: [2, 3], 2: [4], 4: [6], 3: [5], 5: [6], 6: [7], 7: []}
        self.assertCollapsed(d, d)


class TestGraphThunkIdsToKeys(tests.TestCase):

    def test_heads(self):
        # A
        # |\
        # B C
        # |/
        # D
        d = {(b'D',): [(b'B',), (b'C',)], (b'C',): [(b'A',)],
             (b'B',): [(b'A',)], (b'A',): []}
        g = _mod_graph.Graph(_mod_graph.DictParentsProvider(d))
        graph_thunk = _mod_graph.GraphThunkIdsToKeys(g)
        self.assertEqual([b'D'], sorted(graph_thunk.heads([b'D', b'A'])))
        self.assertEqual([b'D'], sorted(graph_thunk.heads([b'D', b'B'])))
        self.assertEqual([b'D'], sorted(graph_thunk.heads([b'D', b'C'])))
        self.assertEqual([b'B', b'C'], sorted(graph_thunk.heads([b'B', b'C'])))

    def test_add_node(self):
        d = {(b'C',): [(b'A',)], (b'B',): [(b'A',)], (b'A',): []}
        g = _mod_graph.KnownGraph(d)
        graph_thunk = _mod_graph.GraphThunkIdsToKeys(g)
        graph_thunk.add_node(b"D", [b"A", b"C"])
        self.assertEqual([b'B', b'D'],
                         sorted(graph_thunk.heads([b'D', b'B', b'A'])))

    def test_merge_sort(self):
        d = {(b'C',): [(b'A',)], (b'B',): [(b'A',)], (b'A',): []}
        g = _mod_graph.KnownGraph(d)
        graph_thunk = _mod_graph.GraphThunkIdsToKeys(g)
        graph_thunk.add_node(b"D", [b"A", b"C"])
        self.assertEqual([(b'C', 0, (2,), False), (b'A', 0, (1,), True)],
                         [(n.key, n.merge_depth, n.revno, n.end_of_merge)
                          for n in graph_thunk.merge_sort(b'C')])


class TestStackedParentsProvider(tests.TestCase):

    def setUp(self):
        super().setUp()
        self.calls = []

    def get_shared_provider(self, info, ancestry, has_cached):
        pp = _mod_graph.DictParentsProvider(ancestry)
        if has_cached:
            pp.get_cached_parent_map = pp.get_parent_map
        return SharedInstrumentedParentsProvider(pp, self.calls, info)

    def test_stacked_parents_provider(self):
        parents1 = _mod_graph.DictParentsProvider({b'rev2': [b'rev3']})
        parents2 = _mod_graph.DictParentsProvider({b'rev1': [b'rev4']})
        stacked = _mod_graph.StackedParentsProvider([parents1, parents2])
        self.assertEqual({b'rev1': [b'rev4'], b'rev2': [b'rev3']},
                         stacked.get_parent_map([b'rev1', b'rev2']))
        self.assertEqual({b'rev2': [b'rev3'], b'rev1': [b'rev4']},
                         stacked.get_parent_map([b'rev2', b'rev1']))
        self.assertEqual({b'rev2': [b'rev3']},
                         stacked.get_parent_map([b'rev2', b'rev2']))
        self.assertEqual({b'rev1': [b'rev4']},
                         stacked.get_parent_map([b'rev1', b'rev1']))

    def test_stacked_parents_provider_overlapping(self):
        # rev2 is availible in both providers.
        # 1
        # |
        # 2
        parents1 = _mod_graph.DictParentsProvider({b'rev2': [b'rev1']})
        parents2 = _mod_graph.DictParentsProvider({b'rev2': [b'rev1']})
        stacked = _mod_graph.StackedParentsProvider([parents1, parents2])
        self.assertEqual({b'rev2': [b'rev1']},
                         stacked.get_parent_map([b'rev2']))

    def test_handles_no_get_cached_parent_map(self):
        # this shows that we both handle when a provider doesn't implement
        # get_cached_parent_map
        pp1 = self.get_shared_provider(b'pp1', {b'rev2': (b'rev1',)},
                                       has_cached=False)
        pp2 = self.get_shared_provider(b'pp2', {b'rev2': (b'rev1',)},
                                       has_cached=True)
        stacked = _mod_graph.StackedParentsProvider([pp1, pp2])
        self.assertEqual({b'rev2': (b'rev1',)},
                         stacked.get_parent_map([b'rev2']))
        # No call on b'pp1' because it doesn't provide get_cached_parent_map
        self.assertEqual([(b'pp2', 'cached', [b'rev2'])], self.calls)

    def test_query_order(self):
        # We should call get_cached_parent_map on all providers before we call
        # get_parent_map. Further, we should track what entries we have found,
        # and not re-try them.
        pp1 = self.get_shared_provider(b'pp1', {b'a': ()}, has_cached=True)
        pp2 = self.get_shared_provider(
            b'pp2', {b'c': (b'b',)}, has_cached=False)
        pp3 = self.get_shared_provider(
            b'pp3', {b'b': (b'a',)}, has_cached=True)
        stacked = _mod_graph.StackedParentsProvider([pp1, pp2, pp3])
        self.assertEqual({b'a': (), b'b': (b'a',), b'c': (b'b',)},
                         stacked.get_parent_map([b'a', b'b', b'c', b'd']))
        self.assertEqual([(b'pp1', 'cached', [b'a', b'b', b'c', b'd']),
                          # No call to pp2, because it doesn't have cached
                          (b'pp3', 'cached', [b'b', b'c', b'd']),
                          (b'pp1', [b'c', b'd']),
                          (b'pp2', [b'c', b'd']),
                          (b'pp3', [b'd']),
                          ], self.calls)



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""General Python convenience functions."""

import sys


def get_named_object(module_name, member_name=None):
    """Get the Python object named by a given module and member name.

    This is usually much more convenient than dealing with ``__import__``
    directly::

        >>> doc = get_named_object('breezy.pyutils', 'get_named_object.__doc__')
        >>> doc.splitlines()[0]
        'Get the Python object named by a given module and member name.'

    :param module_name: a module name, as would be found in sys.modules if
        the module is already imported.  It may contain dots.  e.g. 'sys' or
        'os.path'.
    :param member_name: (optional) a name of an attribute in that module to
        return.  It may contain dots.  e.g. 'MyClass.some_method'.  If not
        given, the named module will be returned instead.
    :raises: ImportError or AttributeError.
    """
    # We may have just a module name, or a module name and a member name,
    # and either may contain dots.  __import__'s return value is a bit
    # unintuitive, so we need to take care to always return the object
    # specified by the full combination of module name + member name.
    if member_name:
        # Give __import__ a from_list.  It will return the last module in
        # the dotted module name.
        attr_chain = member_name.split('.')
        from_list = attr_chain[:1]
        obj = __import__(module_name, {}, {}, from_list)
        for attr in attr_chain:
            obj = getattr(obj, attr)
    else:
        # We're just importing a module, no attributes, so we have no
        # from_list.  __import__ will return the first module in the dotted
        # module name, so we look up the module from sys.modules.
        __import__(module_name, globals(), locals(), [])
        obj = sys.modules[module_name]
    return obj


def calc_parent_name(module_name, member_name=None):
    """Determine the 'parent' of a given dotted module name and (optional)
    member name.

    The idea is that ``getattr(parent_obj, final_attr)`` will equal
    get_named_object(module_name, member_name).

    :return: (module_name, member_name, final_attr) tuple.
    """
# +SKIP is not recognized by python2.4
# Typical use is::
#
#     >>> parent_mod, parent_member, final_attr = calc_parent_name(
#     ...     module_name, member_name) # doctest: +SKIP
#     >>> parent_obj = get_named_object(parent_mod, parent_member)
#     ... # doctest: +SKIP
    if member_name is not None:
        split_name = member_name.rsplit('.', 1)
        if len(split_name) == 1:
            return (module_name, None, member_name)
        else:
            return (module_name, split_name[0], split_name[1])
    else:
        split_name = module_name.rsplit('.', 1)
        if len(split_name) == 1:
            raise AssertionError(
                'No parent object for top-level module {!r}'.format(module_name))
        else:
            return (split_name[0], None, split_name[1])



# Copyright (C) 2010-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests for ref handling."""

from ... import tests
from ..object_store import BazaarObjectStore
from ..refs import BazaarRefsContainer, branch_name_to_ref, ref_to_branch_name


class BranchNameRefConversionTests(tests.TestCase):

    def test_head(self):
        self.assertEqual("", ref_to_branch_name(b"HEAD"))
        self.assertEqual(b"HEAD", branch_name_to_ref(""))

    def test_tag(self):
        self.assertRaises(ValueError, ref_to_branch_name, b"refs/tags/FOO")

    def test_branch(self):
        self.assertEqual("frost", ref_to_branch_name(b"refs/heads/frost"))
        self.assertEqual(b"refs/heads/frost", branch_name_to_ref("frost"))


class BazaarRefsContainerTests(tests.TestCaseWithTransport):

    def test_empty(self):
        tree = self.make_branch_and_tree('.')
        store = BazaarObjectStore(tree.branch.repository)
        refs = BazaarRefsContainer(tree.controldir, store)
        self.assertEqual(refs.as_dict(), {})

    def test_some_commit(self):
        tree = self.make_branch_and_tree('.')
        revid = tree.commit('somechange')
        store = BazaarObjectStore(tree.branch.repository)
        refs = BazaarRefsContainer(tree.controldir, store)
        self.assertEqual(
            refs.as_dict(),
            {b'HEAD': store._lookup_revision_sha1(revid)})

    def test_some_tag(self):
        tree = self.make_branch_and_tree('.')
        revid = tree.commit('somechange')
        tree.branch.tags.set_tag('sometag', revid)
        store = BazaarObjectStore(tree.branch.repository)
        refs = BazaarRefsContainer(tree.controldir, store)
        self.assertEqual(
            refs.as_dict(),
            {b'HEAD': store._lookup_revision_sha1(revid),
             b'refs/tags/sometag': store._lookup_revision_sha1(revid),
             })

    def test_some_branch(self):
        tree = self.make_branch_and_tree('.')
        revid = tree.commit('somechange')
        otherbranch = tree.controldir.create_branch(name='otherbranch')
        otherbranch.generate_revision_history(revid)
        store = BazaarObjectStore(tree.branch.repository)
        refs = BazaarRefsContainer(tree.controldir, store)
        self.assertEqual(
            refs.as_dict(),
            {b'HEAD': store._lookup_revision_sha1(revid),
             b'refs/heads/otherbranch': store._lookup_revision_sha1(revid),
             })



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""WorkingTree3 format and implementation.

"""

import errno

from .. import errors, osutils
from .. import revision as _mod_revision
from .. import trace
from .. import transport as _mod_transport
from ..lockable_files import LockableFiles
from ..lockdir import LockDir
from ..mutabletree import MutableTree
from ..transport.local import LocalTransport
from . import bzrdir, hashcache, inventory
from . import transform as bzr_transform
from .workingtree import InventoryWorkingTree, WorkingTreeFormatMetaDir


class PreDirStateWorkingTree(InventoryWorkingTree):

    def __init__(self, basedir='.', *args, **kwargs):
        super().__init__(basedir, *args, **kwargs)
        # update the whole cache up front and write to disk if anything changed;
        # in the future we might want to do this more selectively
        # two possible ways offer themselves : in self._unlock, write the cache
        # if needed, or, when the cache sees a change, append it to the hash
        # cache file, and have the parser take the most recent entry for a
        # given path only.
        wt_trans = self.controldir.get_workingtree_transport(None)
        cache_filename = wt_trans.local_abspath('stat-cache')
        self._hashcache = hashcache.HashCache(basedir, cache_filename,
                                              self.controldir._get_file_mode(),
                                              self._content_filter_stack_provider())
        hc = self._hashcache
        hc.read()
        # is this scan needed ? it makes things kinda slow.
        # hc.scan()

        if hc.needs_write:
            trace.mutter("write hc")
            hc.write()

    def _write_hashcache_if_dirty(self):
        """Write out the hashcache if it is dirty."""
        if self._hashcache.needs_write:
            try:
                self._hashcache.write()
            except OSError as e:
                if e.errno not in (errno.EPERM, errno.EACCES):
                    raise
                # TODO: jam 20061219 Should this be a warning? A single line
                #       warning might be sufficient to let the user know what
                #       is going on.
                trace.mutter('Could not write hashcache for %s\nError: %s',
                             self._hashcache.cache_file_name(),
                             osutils.safe_unicode(e.args[1]))

    def get_file_sha1(self, path, stat_value=None):
        with self.lock_read():
            # To make sure NoSuchFile gets raised..
            if not self.is_versioned(path):
                raise _mod_transport.NoSuchFile(path)
            return self._hashcache.get_sha1(path, stat_value)


class WorkingTree3(PreDirStateWorkingTree):
    """This is the Format 3 working tree.

    This differs from the base WorkingTree by:
     - having its own file lock
     - having its own last-revision property.

    This is new in bzr 0.8
    """

    def _last_revision(self):
        """See Mutable.last_revision."""
        with self.lock_read():
            try:
                return self._transport.get_bytes('last-revision')
            except _mod_transport.NoSuchFile:
                return _mod_revision.NULL_REVISION

    def _change_last_revision(self, revision_id):
        """See WorkingTree._change_last_revision."""
        if revision_id is None or revision_id == _mod_revision.NULL_REVISION:
            try:
                self._transport.delete('last-revision')
            except _mod_transport.NoSuchFile:
                pass
            return False
        else:
            self._transport.put_bytes('last-revision', revision_id,
                                      mode=self.controldir._get_file_mode())
            return True

    def _get_check_refs(self):
        """Return the references needed to perform a check of this tree."""
        return [('trees', self.last_revision())]

    def unlock(self):
        if self._control_files._lock_count == 1:
            # do non-implementation specific cleanup
            self._cleanup()
            # _inventory_is_modified is always False during a read lock.
            if self._inventory_is_modified:
                self.flush()
            self._write_hashcache_if_dirty()
        # reverse order of locking.
        try:
            return self._control_files.unlock()
        finally:
            self.branch.unlock()


class WorkingTreeFormat3(WorkingTreeFormatMetaDir):
    """The second working tree format updated to record a format marker.

    This format:
        - exists within a metadir controlling .bzr
        - includes an explicit version marker for the workingtree control
          files, separate from the ControlDir format
        - modifies the hash cache format
        - is new in bzr 0.8
        - uses a LockDir to guard access for writes.
    """

    upgrade_recommended = True

    missing_parent_conflicts = True

    supports_versioned_directories = True

    @classmethod
    def get_format_string(cls):
        """See WorkingTreeFormat.get_format_string()."""
        return b"Bazaar-NG Working Tree format 3"

    def get_format_description(self):
        """See WorkingTreeFormat.get_format_description()."""
        return "Working tree format 3"

    _tree_class = WorkingTree3

    def __get_matchingcontroldir(self):
        return bzrdir.BzrDirMetaFormat1()

    _matchingcontroldir = property(__get_matchingcontroldir)

    def _open_control_files(self, a_controldir):
        transport = a_controldir.get_workingtree_transport(None)
        return LockableFiles(transport, 'lock', LockDir)

    def initialize(self, a_controldir, revision_id=None, from_branch=None,
                   accelerator_tree=None, hardlink=False):
        """See WorkingTreeFormat.initialize().

        :param revision_id: if supplied, create a working tree at a different
            revision than the branch is at.
        :param accelerator_tree: A tree which can be used for retrieving file
            contents more quickly than the revision tree, i.e. a workingtree.
            The revision tree will be used for cases where accelerator_tree's
            content is different.
        :param hardlink: If true, hard-link files from accelerator_tree,
            where possible.
        """
        if not isinstance(a_controldir.transport, LocalTransport):
            raise errors.NotLocalUrl(a_controldir.transport.base)
        transport = a_controldir.get_workingtree_transport(self)
        control_files = self._open_control_files(a_controldir)
        control_files.create_lock()
        control_files.lock_write()
        transport.put_bytes('format', self.as_string(),
                            mode=a_controldir._get_file_mode())
        if from_branch is not None:
            branch = from_branch
        else:
            branch = a_controldir.open_branch()
        if revision_id is None:
            revision_id = branch.last_revision()
        # WorkingTree3 can handle an inventory which has a unique root id.
        # as of bzr 0.12. However, bzr 0.11 and earlier fail to handle
        # those trees. And because there isn't a format bump inbetween, we
        # are maintaining compatibility with older clients.
        # inv = Inventory(root_id=gen_root_id())
        inv = self._initial_inventory()
        wt = self._tree_class(a_controldir.root_transport.local_abspath('.'),
                              branch,
                              inv,
                              _internal=True,
                              _format=self,
                              _controldir=a_controldir,
                              _control_files=control_files)
        wt.lock_tree_write()
        try:
            basis_tree = branch.repository.revision_tree(revision_id)
            # only set an explicit root id if there is one to set.
            if basis_tree.path2id('') is not None:
                wt.set_root_id(basis_tree.path2id(''))
            if revision_id == _mod_revision.NULL_REVISION:
                wt.set_parent_trees([])
            else:
                wt.set_parent_trees([(revision_id, basis_tree)])
            bzr_transform.build_tree(basis_tree, wt)
            for hook in MutableTree.hooks['post_build_tree']:
                hook(wt)
        finally:
            # Unlock in this order so that the unlock-triggers-flush in
            # WorkingTree is given a chance to fire.
            control_files.unlock()
            wt.unlock()
        return wt

    def _initial_inventory(self):
        return inventory.Inventory()

    def open(self, a_controldir, _found=False):
        """Return the WorkingTree object for a_controldir

        _found is a private parameter, do not use it. It is used to indicate
               if format probing has already been done.
        """
        if not _found:
            # we are being called directly and must probe.
            raise NotImplementedError
        if not isinstance(a_controldir.transport, LocalTransport):
            raise errors.NotLocalUrl(a_controldir.transport.base)
        wt = self._open(a_controldir, self._open_control_files(a_controldir))
        return wt

    def _open(self, a_controldir, control_files):
        """Open the tree itself.

        :param a_controldir: the dir for the tree.
        :param control_files: the control files for the tree.
        """
        return self._tree_class(a_controldir.root_transport.local_abspath('.'),
                                _internal=True,
                                _format=self,
                                _controldir=a_controldir,
                                _control_files=control_files)



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# The normalize function is taken from pygettext which is distributed
# with Python under the Python License, which is GPL compatible.

"""Extract docstrings from Bazaar commands.

This module only handles breezy objects that use strings not directly wrapped
by a gettext() call. To generate a complete translation template file, this
output needs to be combined with that of xgettext or a similar command for
extracting those strings, as is done in the bzr Makefile. Sorting the output
is also left to that stage of the process.
"""

import inspect
import os
import sys

import breezy

from . import commands as _mod_commands
from . import errors, help_topics, option
from . import plugin as _mod_plugin
from .i18n import gettext
from .trace import mutter, note


def _escape(s):
    s = (s.replace('\\', '\\\\')
         .replace('\n', '\\n')
         .replace('\r', '\\r')
         .replace('\t', '\\t')
         .replace('"', '\\"')
         )
    return s


def _normalize(s):
    # This converts the various Python string types into a format that
    # is appropriate for .po files, namely much closer to C style.
    lines = s.split('\n')
    if len(lines) == 1:
        s = '"' + _escape(s) + '"'
    else:
        if not lines[-1]:
            del lines[-1]
            lines[-1] = lines[-1] + '\n'
        lineterm = '\\n"\n"'
        s = '""\n"' + lineterm.join(map(_escape, lines)) + '"'
    return s


def _parse_source(source_text, filename='<unknown>'):
    """Get object to lineno mappings from given source_text"""
    import ast
    cls_to_lineno = {}
    str_to_lineno = {}
    for node in ast.walk(ast.parse(source_text, filename)):
        # TODO: worry about duplicates?
        if isinstance(node, ast.ClassDef):
            # TODO: worry about nesting?
            cls_to_lineno[node.name] = node.lineno
        elif isinstance(node, ast.Str):
            # Python AST gives location of string literal as the line the
            # string terminates on. It's more useful to have the line the
            # string begins on. Unfortunately, counting back newlines is
            # only an approximation as the AST is ignorant of escaping.
            str_to_lineno[node.s] = node.lineno - (0 if sys.version_info >= (3, 8) else node.s.count('\n'))
    return cls_to_lineno, str_to_lineno


class _ModuleContext:
    """Record of the location within a source tree"""

    def __init__(self, path, lineno=1, _source_info=None):
        self.path = path
        self.lineno = lineno
        if _source_info is not None:
            self._cls_to_lineno, self._str_to_lineno = _source_info

    @classmethod
    def from_module(cls, module):
        """Get new context from module object and parse source for linenos"""
        sourcepath = inspect.getsourcefile(module)
        # TODO: fix this to do the right thing rather than rely on cwd
        relpath = os.path.relpath(sourcepath)
        return cls(relpath,
                   _source_info=_parse_source("".join(inspect.findsource(module)[0]), module.__file__))

    def from_class(self, cls):
        """Get new context with same details but lineno of class in source"""
        try:
            lineno = self._cls_to_lineno[cls.__name__]
        except (AttributeError, KeyError):
            mutter("Definition of %r not found in %r", cls, self.path)
            return self
        return self.__class__(self.path, lineno,
                              (self._cls_to_lineno, self._str_to_lineno))

    def from_string(self, string):
        """Get new context with same details but lineno of string in source"""
        try:
            lineno = self._str_to_lineno[string]
        except (AttributeError, KeyError):
            mutter("String %r not found in %r", string[:20], self.path)
            return self
        return self.__class__(self.path, lineno,
                              (self._cls_to_lineno, self._str_to_lineno))


class _PotExporter:
    """Write message details to output stream in .pot file format"""

    def __init__(self, outf, include_duplicates=False):
        self.outf = outf
        if include_duplicates:
            self._msgids = None
        else:
            self._msgids = set()
        self._module_contexts = {}

    def poentry(self, path, lineno, s, comment=None):
        if self._msgids is not None:
            if s in self._msgids:
                return
            self._msgids.add(s)
        if comment is None:
            comment = ''
        else:
            comment = "# %s\n" % comment
        mutter("Exporting msg %r at line %d in %r", s[:20], lineno, path)
        line = (
            "#: {path}:{lineno}\n"
            "{comment}"
            "msgid {msg}\n"
            "msgstr \"\"\n"
            "\n".format(
                path=path, lineno=lineno, comment=comment, msg=_normalize(s)))
        self.outf.write(line)

    def poentry_in_context(self, context, string, comment=None):
        context = context.from_string(string)
        self.poentry(context.path, context.lineno, string, comment)

    def poentry_per_paragraph(self, path, lineno, msgid, include=None):
        # TODO: How to split long help?
        paragraphs = msgid.split('\n\n')
        if include is not None:
            paragraphs = filter(include, paragraphs)
        for p in paragraphs:
            self.poentry(path, lineno, p)
            lineno += p.count('\n') + 2

    def get_context(self, obj):
        module = inspect.getmodule(obj)
        try:
            context = self._module_contexts[module.__name__]
        except KeyError:
            context = _ModuleContext.from_module(module)
            self._module_contexts[module.__name__] = context
        if inspect.isclass(obj):
            context = context.from_class(obj)
        return context


def _write_option(exporter, context, opt, note):
    if getattr(opt, 'hidden', False):
        return
    optname = opt.name
    if getattr(opt, 'title', None):
        exporter.poentry_in_context(context, opt.title,
                                    f"title of {optname!r} {note}")
    for name, _, _, helptxt in opt.iter_switches():
        if name != optname:
            if opt.is_hidden(name):
                continue
            name = "=".join([optname, name])
        if helptxt:
            exporter.poentry_in_context(context, helptxt,
                                        f"help of {name!r} {note}")


def _standard_options(exporter):
    OPTIONS = option.Option.OPTIONS
    context = exporter.get_context(option)
    for name in sorted(OPTIONS):
        opt = OPTIONS[name]
        _write_option(exporter, context.from_string(name), opt, "option")


def _command_options(exporter, context, cmd):
    note = f"option of {cmd.name()!r} command"
    for opt in cmd.takes_options:
        # String values in Command option lists are for global options
        if not isinstance(opt, str):
            _write_option(exporter, context, opt, note)


def _write_command_help(exporter, cmd):
    context = exporter.get_context(cmd.__class__)
    rawdoc = cmd.__doc__
    dcontext = context.from_string(rawdoc)
    doc = inspect.cleandoc(rawdoc)

    def exclude_usage(p):
        # ':Usage:' has special meaning in help topics.
        # This is usage example of command and should not be translated.
        if p.splitlines()[0] != ':Usage:':
            return True

    exporter.poentry_per_paragraph(dcontext.path, dcontext.lineno, doc,
                                   exclude_usage)
    _command_options(exporter, context, cmd)


def _command_helps(exporter, plugin_name=None):
    """Extract docstrings from path.

    This respects the Bazaar cmdtable/table convention and will
    only extract docstrings from functions mentioned in these tables.
    """

    # builtin commands
    for cmd_name in _mod_commands.builtin_command_names():
        command = _mod_commands.get_cmd_object(cmd_name, False)
        if command.hidden:
            continue
        if plugin_name is not None:
            # only export builtins if we are not exporting plugin commands
            continue
        note(gettext("Exporting messages from builtin command: %s"), cmd_name)
        _write_command_help(exporter, command)

    plugins = _mod_plugin.plugins()
    if plugin_name is not None and plugin_name not in plugins:
        raise errors.BzrError(gettext('Plugin %s is not loaded' % plugin_name))
    core_plugins = {
        name for name in plugins
        if plugins[name].path().startswith(breezy.__path__[0])}
    # plugins
    for cmd_name in _mod_commands.plugin_command_names():
        command = _mod_commands.get_cmd_object(cmd_name, False)
        if command.hidden:
            continue
        if plugin_name is not None and command.plugin_name() != plugin_name:
            # if we are exporting plugin commands, skip plugins we have not
            # specified.
            continue
        if plugin_name is None and command.plugin_name() not in core_plugins:
            # skip non-core plugins
            # TODO: Support extracting from third party plugins.
            continue
        note(gettext("Exporting messages from plugin command: {0} in {1}").format(
             cmd_name, command.plugin_name()))
        _write_command_help(exporter, command)


def _error_messages(exporter):
    """Extract fmt string from breezy.errors."""
    context = exporter.get_context(errors)
    base_klass = errors.BzrError
    for name in dir(errors):
        klass = getattr(errors, name)
        if not inspect.isclass(klass):
            continue
        if not issubclass(klass, base_klass):
            continue
        if klass is base_klass:
            continue
        if klass.internal_error:
            continue
        fmt = getattr(klass, "_fmt", None)
        if fmt:
            note(gettext("Exporting message from error: %s"), name)
            exporter.poentry_in_context(context, fmt)


def _help_topics(exporter):
    topic_registry = help_topics.topic_registry
    for key in topic_registry.keys():
        doc = topic_registry.get(key)
        if isinstance(doc, str):
            exporter.poentry_per_paragraph(
                'dummy/help_topics/' + key + '/detail.txt',
                1, doc)
        elif callable(doc):  # help topics from files
            exporter.poentry_per_paragraph(
                'en/help_topics/' + key + '.txt',
                1, doc(key))
        summary = topic_registry.get_summary(key)
        if summary is not None:
            exporter.poentry('dummy/help_topics/' + key + '/summary.txt',
                             1, summary)


def export_pot(outf, plugin=None, include_duplicates=False):
    exporter = _PotExporter(outf, include_duplicates)
    if plugin is None:
        _standard_options(exporter)
        _command_helps(exporter)
        _error_messages(exporter)
        _help_topics(exporter)
    else:
        _command_helps(exporter, plugin)



# Copyright (C) 2005, 2006, 2008-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test for setup.py build process"""

import os
import subprocess
import sys

from packaging.version import Version

import breezy

from .. import tests

# TODO: Run bzr from the installed copy to see if it works.  Really we need to
# run something that exercises every module, just starting it may not detect
# some missing modules.
#
# TODO: Check that the version numbers are in sync.  (Or avoid this...)


class TestSetup(tests.TestCaseInTempDir):

    def test_build_and_install(self):
        """ test cmd `python setup.py build`

        This tests that the build process and man generator run correctly.
        It also can catch new subdirectories that weren't added to setup.py.
        """
        # setup.py must be run from the root source directory, but the tests
        # are not necessarily invoked from there
        self.source_dir = os.path.dirname(os.path.dirname(breezy.__file__))
        if not os.path.isfile(os.path.join(self.source_dir, 'setup.py')):
            self.skipTest(
                'There is no setup.py file adjacent to the breezy directory')
        if os.environ.get('GITHUB_ACTIONS') == 'true':
            # On GitHub CI, for some reason rustc can't be found.
            # Marking as known failing for now.
            self.knownFailure(
                'rustc can not be found in the GitHub actions environment')
        try:
            import distutils.sysconfig
            makefile_path = distutils.sysconfig.get_makefile_filename()
            if not os.path.exists(makefile_path):
                self.skipTest(
                    'You must have the python Makefile installed to run this'
                    ' test. Usually this can be found by installing'
                    ' "python-dev"')
        except ImportError:
            self.skipTest(
                'You must have distutils installed to run this test.'
                ' Usually this can be found by installing "python-dev"')
        self.log('test_build running from %s' % self.source_dir)
        build_dir = os.path.join(self.test_dir, "build")
        install_dir = os.path.join(self.test_dir, "install")
        self.run_setup([
            'build', '-b', build_dir,
            'install', '--root', install_dir])
        # Install layout is platform dependant
        self.assertPathExists(install_dir)
        self.run_setup(['clean', '-b', build_dir])

    def run_setup(self, args):
        args = [sys.executable, './setup.py', ] + args
        self.log('source base directory: %s', self.source_dir)
        self.log('args: %r', args)
        env = dict(os.environ)
        env['PYTHONPATH'] = ':'.join(sys.path)
        p = subprocess.Popen(args,
                             cwd=self.source_dir,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             env=env)
        stdout, stderr = p.communicate()
        self.log('stdout: %r', stdout)
        self.log('stderr: %r', stderr)
        self.assertEqual(0, p.returncode,
                         'invocation of %r failed' % args)


class TestDistutilsVersion(tests.TestCase):

    def test_version_with_string(self):
        # We really care about two pyrex specific versions and our ability to
        # detect them
        lv = Version
        self.assertTrue(lv("0.9.4.1") < lv('0.17.beta1'))
        self.assertTrue(lv("0.9.6.3") < lv('0.10'))



# Copyright (C) 2005, 2006, 2008, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that various operations work in a non-ASCII environment."""

import os
import sys
from unicodedata import normalize

from .. import osutils
from ..osutils import pathjoin
from . import TestCase, TestCaseWithTransport, TestSkipped


class NonAsciiTest(TestCaseWithTransport):

    def test_add_in_nonascii_branch(self):
        """Test adding in a non-ASCII branch."""
        br_dir = "\u1234"
        try:
            wt = self.make_branch_and_tree(br_dir)
        except UnicodeEncodeError:
            raise TestSkipped("filesystem can't accomodate nonascii names")
            return
        with open(pathjoin(br_dir, "a"), "w") as f:
            f.write("hello")
        wt.add(["a"], ids=[b"a-id"])


a_circle_c = '\xe5'
a_circle_d = 'a\u030a'
a_dots_c = '\xe4'
a_dots_d = 'a\u0308'
z_umlat_c = '\u017d'
z_umlat_d = 'Z\u030c'
squared_c = '\xbc'  # This gets mapped to '2' if we use NFK[CD]
squared_d = '\xbc'
quarter_c = '\xb2'  # Gets mapped to u'1\u20444' (1/4) if we use NFK[CD]
quarter_d = '\xb2'


class TestNormalization(TestCase):
    """Verify that we have our normalizations correct."""

    def test_normalize(self):
        self.assertEqual(a_circle_d, normalize('NFD', a_circle_c))
        self.assertEqual(a_circle_c, normalize('NFC', a_circle_d))
        self.assertEqual(a_dots_d, normalize('NFD', a_dots_c))
        self.assertEqual(a_dots_c, normalize('NFC', a_dots_d))
        self.assertEqual(z_umlat_d, normalize('NFD', z_umlat_c))
        self.assertEqual(z_umlat_c, normalize('NFC', z_umlat_d))
        self.assertEqual(squared_d, normalize('NFC', squared_c))
        self.assertEqual(squared_c, normalize('NFD', squared_d))
        self.assertEqual(quarter_d, normalize('NFC', quarter_c))
        self.assertEqual(quarter_c, normalize('NFD', quarter_d))


class NormalizedFilename(TestCaseWithTransport):
    """Test normalized_filename and associated helpers"""

    def test__accessible_normalized_filename(self):
        anf = osutils._accessible_normalized_filename
        # normalized_filename should allow plain ascii strings
        # not just unicode strings
        self.assertEqual(('ascii', True), anf('ascii'))
        self.assertEqual((a_circle_c, True), anf(a_circle_c))
        self.assertEqual((a_circle_c, True), anf(a_circle_d))
        self.assertEqual((a_dots_c, True), anf(a_dots_c))
        self.assertEqual((a_dots_c, True), anf(a_dots_d))
        self.assertEqual((z_umlat_c, True), anf(z_umlat_c))
        self.assertEqual((z_umlat_c, True), anf(z_umlat_d))
        self.assertEqual((squared_c, True), anf(squared_c))
        self.assertEqual((squared_c, True), anf(squared_d))
        self.assertEqual((quarter_c, True), anf(quarter_c))
        self.assertEqual((quarter_c, True), anf(quarter_d))

    def test__inaccessible_normalized_filename(self):
        inf = osutils._inaccessible_normalized_filename
        # normalized_filename should allow plain ascii strings
        # not just unicode strings
        self.assertEqual(('ascii', True), inf('ascii'))
        self.assertEqual((a_circle_c, True), inf(a_circle_c))
        self.assertEqual((a_circle_c, False), inf(a_circle_d))
        self.assertEqual((a_dots_c, True), inf(a_dots_c))
        self.assertEqual((a_dots_c, False), inf(a_dots_d))
        self.assertEqual((z_umlat_c, True), inf(z_umlat_c))
        self.assertEqual((z_umlat_c, False), inf(z_umlat_d))
        self.assertEqual((squared_c, True), inf(squared_c))
        self.assertEqual((squared_c, True), inf(squared_d))
        self.assertEqual((quarter_c, True), inf(quarter_c))
        self.assertEqual((quarter_c, True), inf(quarter_d))

    def test_functions(self):
        if osutils.normalizes_filenames():
            self.assertEqual(osutils.normalized_filename,
                             osutils._accessible_normalized_filename)
        else:
            self.assertEqual(osutils.normalized_filename,
                             osutils._inaccessible_normalized_filename)

    def test_platform(self):
        # With FAT32 and certain encodings on win32
        # a_circle_c and a_dots_c actually map to the same file
        # adding a suffix kicks in the 'preserving but insensitive'
        # route, and maintains the right files
        files = [a_circle_c + '.1', a_dots_c + '.2', z_umlat_c + '.3']
        try:
            self.build_tree(files)
        except UnicodeError:
            raise TestSkipped("filesystem cannot create unicode files")

        if sys.platform == 'darwin':
            expected = sorted(
                [a_circle_d + '.1', a_dots_d + '.2', z_umlat_d + '.3'])
        else:
            expected = sorted(files)

        present = sorted(os.listdir('.'))
        self.assertEqual(expected, present)

    def test_access_normalized(self):
        # We should always be able to access files created with
        # normalized filenames
        # With FAT32 and certain encodings on win32
        # a_circle_c and a_dots_c actually map to the same file
        # adding a suffix kicks in the 'preserving but insensitive'
        # route, and maintains the right files
        files = [a_circle_c + '.1', a_dots_c + '.2', z_umlat_c + '.3',
                 squared_c + '.4', quarter_c + '.5']
        try:
            self.build_tree(files, line_endings='native')
        except UnicodeError:
            raise TestSkipped("filesystem cannot create unicode files")

        for fname in files:
            # We should get an exception if we can't open the file at
            # this location.
            path, can_access = osutils.normalized_filename(fname)

            self.assertEqual(path, fname)
            self.assertTrue(can_access)

            with open(path, 'rb') as f:
                # Check the contents
                shouldbe = b'contents of %s%s' % (path.encode('utf8'),
                                                  os.linesep.encode('utf-8'))
                actual = f.read()
            self.assertEqual(shouldbe, actual,
                             'contents of %r is incorrect: %r != %r'
                             % (path, shouldbe, actual))

    def test_access_non_normalized(self):
        # Sometimes we can access non-normalized files by their normalized
        # path, verify that normalized_filename returns the right info
        files = [a_circle_d + '.1', a_dots_d + '.2', z_umlat_d + '.3']

        try:
            self.build_tree(files)
        except UnicodeError:
            raise TestSkipped("filesystem cannot create unicode files")

        for fname in files:
            # We should get an exception if we can't open the file at
            # this location.
            path, can_access = osutils.normalized_filename(fname)

            self.assertNotEqual(path, fname)

            # We should always be able to access them from the name
            # they were created with
            f = open(fname, 'rb')
            f.close()

            # And normalized_filename sholud tell us correctly if we can
            # access them by an alternate name
            if can_access:
                f = open(path, 'rb')
                f.close()
            else:
                self.assertRaises(IOError, open, path, 'rb')



# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Commit message editor support."""

import codecs
import os
import sys
from io import BytesIO, StringIO
from subprocess import call

from . import bedding, cmdline, config, osutils, trace, transport, ui
from .errors import BzrError
from .hooks import Hooks


class BadCommitMessageEncoding(BzrError):

    _fmt = 'The specified commit message contains characters unsupported by '\
        'the current encoding.'


def _get_editor():
    """Return sequence of possible editor binaries for the current platform"""
    try:
        yield os.environ["BRZ_EDITOR"], '$BRZ_EDITOR'
    except KeyError:
        pass

    e = config.GlobalStack().get('editor')
    if e is not None:
        yield e, bedding.config_path()

    for varname in 'VISUAL', 'EDITOR':
        if varname in os.environ:
            yield os.environ[varname], '$' + varname

    if sys.platform == 'win32':
        for editor in 'wordpad.exe', 'notepad.exe':
            yield editor, None
    else:
        for editor in ['/usr/bin/editor', 'vi', 'pico', 'nano', 'joe']:
            yield editor, None


def _run_editor(filename):
    """Try to execute an editor to edit the commit message."""
    for candidate, candidate_source in _get_editor():
        edargs = cmdline.split(candidate)
        try:
            x = call(edargs + [filename])
        except OSError as e:
            if candidate_source is not None:
                # We tried this editor because some user configuration (an
                # environment variable or config file) said to try it.  Let
                # the user know their configuration is broken.
                trace.warning(
                    'Could not start editor "%s" (specified by %s): %s\n'
                    % (candidate, candidate_source, str(e)))
            continue
            raise
        if x == 0:
            return True
        elif x == 127:
            continue
        else:
            break
    raise BzrError("Could not start any editor.\nPlease specify one with:\n"
                   " - $BRZ_EDITOR\n - editor=/some/path in %s\n"
                   " - $VISUAL\n - $EDITOR" %
                   bedding.config_path())


DEFAULT_IGNORE_LINE = "%(bar)s %(msg)s %(bar)s" % \
    {'bar': '-' * 14, 'msg': 'This line and the following will be ignored'}


def edit_commit_message(infotext, ignoreline=DEFAULT_IGNORE_LINE,
                        start_message=None):
    """Let the user edit a commit message in a temp file.

    This is run if they don't give a message or
    message-containing file on the command line.

    :param infotext:    Text to be displayed at bottom of message
                        for the user's reference;
                        currently similar to 'bzr status'.

    :param ignoreline:  The separator to use above the infotext.

    :param start_message:   The text to place above the separator, if any.
                            This will not be removed from the message
                            after the user has edited it.

    :return:    commit message or None.
    """

    if start_message is not None:
        start_message = start_message.encode(osutils.get_user_encoding())
    infotext = infotext.encode(osutils.get_user_encoding(), 'replace')
    return edit_commit_message_encoded(infotext, ignoreline, start_message)


def edit_commit_message_encoded(infotext, ignoreline=DEFAULT_IGNORE_LINE,
                                start_message=None):
    """Let the user edit a commit message in a temp file.

    This is run if they don't give a message or
    message-containing file on the command line.

    :param infotext:    Text to be displayed at bottom of message
                        for the user's reference;
                        currently similar to 'bzr status'.
                        The string is already encoded

    :param ignoreline:  The separator to use above the infotext.

    :param start_message:   The text to place above the separator, if any.
                            This will not be removed from the message
                            after the user has edited it.
                            The string is already encoded

    :return:    commit message or None.
    """
    msgfilename = None
    try:
        msgfilename, hasinfo = _create_temp_file_with_commit_template(
            infotext, ignoreline, start_message)
        if not msgfilename:
            return None
        basename = osutils.basename(msgfilename)
        msg_transport = transport.get_transport_from_path(
            osutils.dirname(msgfilename))
        reference_content = msg_transport.get_bytes(basename)
        if not _run_editor(msgfilename):
            return None
        edited_content = msg_transport.get_bytes(basename)
        if edited_content == reference_content:
            if not ui.ui_factory.confirm_action(
                "Commit message was not edited, use anyway",
                "breezy.msgeditor.unchanged",
                    {}):
                # Returning "" makes cmd_commit raise 'empty commit message
                # specified' which is a reasonable error, given the user has
                # rejected using the unedited template.
                return ""
        started = False
        msg = []
        lastline, nlines = 0, 0
        with codecs.open(msgfilename, mode='rb', encoding=osutils.get_user_encoding()) as f:
            try:
                for line in f:
                    stripped_line = line.strip()
                    # strip empty line before the log message starts
                    if not started:
                        if stripped_line != "":
                            started = True
                        else:
                            continue
                    # check for the ignore line only if there
                    # is additional information at the end
                    if hasinfo and stripped_line == ignoreline:
                        break
                    nlines += 1
                    # keep track of the last line that had some content
                    if stripped_line != "":
                        lastline = nlines
                    msg.append(line)
            except UnicodeDecodeError:
                raise BadCommitMessageEncoding()

        if len(msg) == 0:
            return ""
        # delete empty lines at the end
        del msg[lastline:]
        # add a newline at the end, if needed
        if not msg[-1].endswith("\n"):
            return "{}{}".format("".join(msg), "\n")
        else:
            return "".join(msg)
    finally:
        # delete the msg file in any case
        if msgfilename is not None:
            try:
                os.unlink(msgfilename)
            except OSError as e:
                trace.warning(
                    "failed to unlink %s: %s; ignored", msgfilename, e)


def _create_temp_file_with_commit_template(infotext,
                                           ignoreline=DEFAULT_IGNORE_LINE,
                                           start_message=None,
                                           tmpdir=None):
    """Create temp file and write commit template in it.

    :param infotext: Text to be displayed at bottom of message for the
        user's reference; currently similar to 'bzr status'.  The text is
        already encoded.

    :param ignoreline:  The separator to use above the infotext.

    :param start_message: The text to place above the separator, if any.
        This will not be removed from the message after the user has edited
        it.  The string is already encoded

    :return:    2-tuple (temp file name, hasinfo)
    """
    import tempfile
    tmp_fileno, msgfilename = tempfile.mkstemp(prefix='bzr_log.',
                                               dir=tmpdir, text=True)
    with os.fdopen(tmp_fileno, 'wb') as msgfile:
        if start_message is not None:
            msgfile.write(b"%s\n" % start_message)

        if infotext is not None and infotext != "":
            hasinfo = True
            trailer = b"\n\n%s\n\n%s" % (
                ignoreline.encode(osutils.get_user_encoding()), infotext)
            msgfile.write(trailer)
        else:
            hasinfo = False

    return (msgfilename, hasinfo)


def make_commit_message_template(working_tree, specific_files):
    """Prepare a template file for a commit into a branch.

    Returns a unicode string containing the template.
    """
    # TODO: make provision for this to be overridden or modified by a hook
    #
    # TODO: Rather than running the status command, should prepare a draft of
    # the revision to be committed, then pause and ask the user to
    # confirm/write a message.
    from .status import show_tree_status
    status_tmp = StringIO()
    show_tree_status(working_tree, specific_files=specific_files,
                     to_file=status_tmp, verbose=True)
    return status_tmp.getvalue()


def make_commit_message_template_encoded(working_tree, specific_files,
                                         diff=None, output_encoding='utf-8'):
    """Prepare a template file for a commit into a branch.

    Returns an encoded string.
    """
    # TODO: make provision for this to be overridden or modified by a hook
    #
    # TODO: Rather than running the status command, should prepare a draft of
    # the revision to be committed, then pause and ask the user to
    # confirm/write a message.
    from .diff import show_diff_trees

    template = make_commit_message_template(working_tree, specific_files)
    template = template.encode(output_encoding, "replace")

    if diff:
        stream = BytesIO()
        show_diff_trees(working_tree.basis_tree(),
                        working_tree, stream, specific_files,
                        path_encoding=output_encoding)
        template = template + b'\n' + stream.getvalue()

    return template


class MessageEditorHooks(Hooks):
    """A dictionary mapping hook name to a list of callables for message editor
    hooks.

    e.g. ['commit_message_template'] is the list of items to be called to
    generate a commit message template
    """

    def __init__(self):
        """Create the default hooks.

        These are all empty initially.
        """
        Hooks.__init__(self, "breezy.msgeditor", "hooks")
        self.add_hook(
            'set_commit_message',
            "Set a fixed commit message. "
            "set_commit_message is called with the "
            "breezy.commit.Commit object (so you can also change e.g. "
            "revision properties by editing commit.builder._revprops) and the "
            "message so far. set_commit_message must return the message to "
            "use or None if it should use the message editor as normal.",
            (2, 4))
        self.add_hook(
            'commit_message_template',
            "Called when a commit message is being generated. "
            "commit_message_template is called with the breezy.commit.Commit "
            "object and the message that is known so far. "
            "commit_message_template must return a new message to use (which "
            "could be the same as it was given). When there are multiple "
            "hooks registered for commit_message_template, they are chained "
            "with the result from the first passed into the second, and so "
            "on.", (1, 10))


hooks = MessageEditorHooks()


def set_commit_message(commit, start_message=None):
    """Sets the commit message.
    :param commit: Commit object for the active commit.
    :return: The commit message or None to continue using the message editor
    """
    start_message = None
    for hook in hooks['set_commit_message']:
        start_message = hook(commit, start_message)
    return start_message


def generate_commit_message_template(commit, start_message=None):
    """Generate a commit message template.

    :param commit: Commit object for the active commit.
    :param start_message: Message to start with.
    :return: A start commit message or None for an empty start commit message.
    """
    start_message = None
    for hook in hooks['commit_message_template']:
        start_message = hook(commit, start_message)
    return start_message
